from qepy import *
import sys
import argparse
"""
In this example we construct supercells (sc) of bulk hBN. 
Starting from a unit cell (uc) input that is read from file, we build:

- diagonal supercell
- non-diagonal supercell folding a specific q-point Q
- displaced non-diagonal supercells along phonon eigenmodes at Q (read from file)
"""

def generate_diagonal_supercell(uc,R):
    """
    First case: diagonal supercell of size R
    """
    qe = PwIn.from_file(uc) #read the uc input

    sc = Supercell(qe) # initialize class
    sc.d_sup(R) # Generate supercell as PwIn object called 'qe_d'

    #name of output file
    suffix = ''.join(sc.qe_d.control['calculation'].split('\''))
    prefix = ''.join(sc.qe_d.control['prefix'].split('\''))

    #write supercell to file
    sc.qe_d.write('sc_diagonal_%s.%s'%(prefix,suffix))

    print('Diagonal supercell written to file.')

def generate_nondiagonal_supercell(uc,Q,kpoints=None):
    """
    Second case: nondiagonal supercell folding point Q
    """
    qe = PwIn.from_file(uc) #read the uc input
    if kpoints is not None: qe.kpoints = kpoints #Optional: manually change the original kpt mesh to ensure consistency    

    sc = Supercell(qe) # initialize class
    sc.nd_sup(Q) # Generate supercell as PwIn object called 'qe_nd'

    #name of output file
    suffix = ''.join(sc.qe_nd.control['calculation'].split('\''))
    prefix = ''.join(sc.qe_nd.control['prefix'].split('\''))

    #write supercell to file
    sc.qe_nd.write('sc_nondiagonal_%s.%s'%(prefix,suffix))

    print('Nondiagonal supercell written to file.')   

def generate_displaced_supercells(uc,Q,modes_file,kpoints=None):
    """
    Third case: displaced supercells along phonon modes at Q
    """
    qe = PwIn.from_file(uc) #read the uc input
    if kpoints is not None: qe.kpoints = kpoints #Optional: manually change the original kpt mesh to ensure consistency

    sc = Supercell(qe) # initialize class
    nd_atom_positions = sc.nd_sup(Q) # Generate supercell as PwIn object called 'qe_nd' getting new atomic positions
    
    #Read the matdyn file
    qe_dyn=Matdyn.from_modes_file(filename=modes_file)

    # Displace atoms./
    # Intensity is Temp (in bohr)
    # Sign and direction (standing wave at Q) given by modes_file
    #sc.displace(modes_file,nd_atom_positions,Temp=0.1) # Generate list of displaced supercells as PwIn objects called 'modes_qe'
    # iq index of the q-point in the matdyn file (default 0, the first q-point)
    sc.displace(qe_dyn,nd_atom_positions,iq=0,Temp=0.1)
    N_modes = len(sc.modes_qe)

    #name of output file
    suffix = ''.join(sc.qe_nd.control['calculation'].split('\''))
    prefix = ''.join(sc.qe_nd.control['prefix'].split('\''))

    #write supercells to file for each mode
    for mode in range(N_modes): sc.modes_qe[mode].write('sc_displaced_%s_mode%d.%s'%(prefix,mode+1,suffix)) 

    print('Displaced supercells written to file.')

def generate_displaced_unitcell(uc,modes_file):
    """
    Fourth case: displaced cell at Q=0
    """
    qe = PwIn.from_file(uc) #read the uc input

    sc = Supercell(qe) # initialize class
    atom_positions = sc.d_sup([1,1,1]) # Generate "supercell" with size 1 as PwIn object called 'qe_d' getting atomic positions
    
    #Read the matdyn file
    qe_dyn=Matdyn.from_modes_file(filename=modes_file)

    # Displace atoms.
    # Intensity is Temp (in bohr)
    # Sign and direction (standing wave at Q) given by modes_file
    sc.displace(qe_dyn,atom_positions,Temp=0.1) # Generate list of displaced supercells as PwIn objects called 'modes_qe'
    N_modes = len(sc.modes_qe)

    #name of output file
    suffix = ''.join(sc.qe_d.control['calculation'].split('\''))
    prefix = ''.join(sc.qe_d.control['prefix'].split('\''))

    #write supercells to file for each mode
    for mode in range(N_modes): sc.modes_qe[mode].write('uc_GAMMA_displaced_%s_mode%d.%s'%(prefix,mode+4,suffix))

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Supercell generation')
    parser.add_argument('-D', '--diagonal', action="store_true", help='Build diagonal supercell')
    parser.add_argument('-N', '--nondiagonal', action="store_true", help='Build nondiagonal supercell')
    parser.add_argument('-S', '--displaced_nondiagonal', action="store_true", help='Build displaced nondiagonal supercells')
    parser.add_argument('-G', '--displaced_gamma', action="store_true", help='Build displaced unit cell')
    args = parser.parse_args()

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    uc_filnm = 'uc.scf' #pw input
    R = [3,3,1] #diagonal supercell dimensions
    Q = [[1,-1,0],[3,6,2]] #qpoint to fold in nondiagonal supercell: Q=(1/3,-1/6,0)
    kpoints = [12,12,4]    #NB: these fractional crystal coordinates must exactly divide the kpoint mesh!
    modes_file = 'Q.modes' #file with phonon eigenmodes at Q. NB2: we use those already scaled with the atomic masses

    if args.diagonal: generate_diagonal_supercell(uc_filnm,R)

    if args.nondiagonal: generate_nondiagonal_supercell(uc_filnm,Q,kpoints)

    #NB3: If computing the eigenmodes at Q with qe, remember to use Q in *Cartesian* coordinates
    #     in both qe input and output!!
    #     For example, Q_crystal=(1/3,-1/6,0) ====> Q_cartesian = (1/3,0,0) in the BN hexagonal lattice

    if args.displaced_nondiagonal: generate_displaced_supercells(uc_filnm,Q,modes_file,kpoints)
    
    #NB4: For the following we are using the same modes_file as in the case above for simplicity, but it is *wrong*:
    #     Phonon modes at Gamma must be used in this case.
    if args.displaced_gamma: generate_displaced_unitcell(uc_filnm,modes_file)



"""
Simple script to plot and analyse 
exciton-phonon matrix elements calculated with yambo_ph
"""
import numpy as np
import matplotlib.pyplot as plt
from yambopy import YamboLatticeDB,YamboExcitonPhononDB
from yambopy.units import ha2ev,bohr2ang

def plot2D(excph,exc_in,exc_out,ph_in,plt_cbar=False,**kwargs):
    """ 
    Input indices are lists (even of one element) or 'all'
    
    Example: we can sum |G|^2 over all outgoing states and phonon modes 
             for the first bright exc_in state
    
    np.sum( X.excph_sq, axis=3 ) # sum over exc_out
    np.sum( data, axis=1 )    # sum over phonon modes
    np.sqrt(data[:,2]+data[:,3]) # sum over degenerate states (MoS2 A-bright exc_in)

    We construct F_q to plot
    """

    G_squared = excph.excph_sq
    G2plt = np.zeros(len(G_squared))

    if exc_in  == 'all': exc_in  = range(G_squared.shape[2])
    if exc_out == 'all': exc_out = range(G_squared.shape[3])
    if ph_in   == 'all': ph_in   = range(G_squared.shape[1])

    G_squared = G_squared[:, ph_in, :, :].sum(axis=(1))
    G_squared = G_squared[:, exc_in, :].sum(axis=(1))
    G_squared = G_squared[:, exc_out].sum(axis=(1))

    F_q = np.sqrt( G_squared )*ha2ev # Switch from Ha to eV

    # Do plot
    excph.plot_excph(F_q,plt_cbar=plt_cbar,**kwargs)
    lim = np.linalg.norm(excph.rlat[1])
    lim = 0.85*np.linalg.norm(excph.rlat[0])
    excph.ax.set_xticks([])
    excph.ax.set_yticks([])
    ticks = [-0.15,-0.1,-0.05,0,0.05,0.1,0.15]
    ticklabels = [ "{:.1f}".format(tick*2.*np.pi/bohr2ang) for tick in ticks ]
    excph.ax.set_xticks(ticks)
    excph.ax.set_yticks(ticks)
    excph.ax.set_xticklabels(ticklabels)
    excph.ax.set_yticklabels(ticklabels)
    excph.ax.set_xlim(-lim,lim)
    excph.ax.set_ylim(-lim,lim)
    excph.ax.set_xlabel('$q_x$ ($\mathring{A}^{-1}$)',fontsize=12)
    excph.ax.set_ylabel('$q_y$ ($\mathring{A}^{-1}$)',fontsize=12)
    excph.cbar.ax.set_ylabel('$|F(q)|$ eV',fontsize=12)

#
# User input
#

# Exciton in states
exc_in  = [2,3]     # A: 2,3 -- B: 6,7
exc_out = [0,1,2,3] # first 4 states (dispersion of triplet state and A)
ph_in  = 'all' 

# Paths of databases
path='1L_MoS2'
ns_db1 =f'{path}/SAVE/ns.db1'
ndb_exc=f'{path}/excph'

# Plot details
plt_fl='MoS2_A_exc_2Dmap.png'
plt_title='A exciton'
size = 700

#
# Load data
#

# Read lattice and k-space info
ylat = YamboLatticeDB.from_db_file(filename=ns_db1)
print(ylat)

# Read exc-ph databases
X = YamboExcitonPhononDB(ylat,save_excph=ndb_exc)
print(X)

#
# Plot
#
plot2D(X,exc_in,exc_out,ph_in,plt_cbar=True,\
       marker='H',s=size,cmap='magma')
plt.title(plt_title,fontsize=8)
plt.savefig(plt_fl,dpi=200)
plt.show()



#
# Author: Henrique Pereira Coutada Miranda
# Tests for the yambopy library
# scheduler
#
from __future__ import print_function
import unittest
import sys
import os
import argparse
import subprocess
import filecmp
import json
from schedulerpy import *
from qepy import *
from textwrap import dedent

#name of the configuration file to be created when running the tests
_pre_run_pbs  = dedent( """
                        exec > ${PBS_O_WORKDIR}/${PBS_JOBNAME}_${PBS_JOBID}.log
                        echo "------------------ Work dir --------------------"
                        cd ${PBS_O_WORKDIR} && echo ${PBS_O_WORKDIR}
                        echo "------------------ Job Info --------------------"
                        echo "jobid : $PBS_JOBID"
                        echo "jobname : $PBS_JOBNAME"
                        echo "job type : $PBS_ENVIRONMENT"
                        echo "submit dir : $PBS_O_WORKDIR"
                        echo "queue : $PBS_O_QUEUE"
                        echo "user : $PBS_O_LOGNAME"
                        echo "threads : $OMP_NUM_THREADS"
                        """)
_pre_run_pbs_file = "pre_run_pbs.sh"
_test_config_file = "testconfig.json"
_test_config =  { "default":"bash",
                  "bash":
                    {
                        "mpirun": "mpirun",
                        "modules": "None",
                        "pos_run": ["echo \"done!\""],
                        "modules": {"abinit"  :"abinit/8.0",
                                    "espresso":"espresso/5.4.0"}
                    },
                  "oar":
                    {
                        "mpirun": "mpirun",
                        "pos_run": ["echo \"done!\""],
                        "idempotent": "true",
                        "modules": {"abinit"  :"abinit/8.0",
                                    "espresso":"espresso/5.4.0"}
                    },
                  "slurm":
                    {
                        "mpirun": "mpirun",
                        "pos_run": ["echo \"done!\""],
                        "modules": {"abinit"  :"abinit/8.0",
                                    "espresso":"espresso/5.4.0"}
                    },
                  "pbs":
                    {
                        "mpirun": "mpirun",
                        "pre_run": "file:%s"%_pre_run_pbs_file,
                        "pos_run": ["qstat -f $PBS_JOBID"],
                        "queue": "myqueue",
                        "group_list": "mygroup",
                        "mem": "2624",
                        "var_cores": "ppn",
                        "var_nodes": "select",
                        "modules": {"abinit"  :"abinit/8.0",
                                    "espresso":"espresso/5.4.0"}
                    }
                }
_test_file = "test.sh"

def header(text):
    return "\n%s\n%s\n%s\n"%("="*10,text,"="*10)

def init_config():
    """
    create basic configuration file
    """
    #write pbs pre_run example
    f = open(_pre_run_pbs_file,'w')
    f.write(_pre_run_pbs)
    f.close()
    #write test configuration file
    f = open(_test_config_file,'w')
    json.dump(_test_config,f)
    f.close()
    #scheduler
    Scheduler._config_path     = os.getcwd()
    Scheduler._config_filename = _test_config_file

def clean_config():
    os.remove(_pre_run_pbs_file)
    os.remove(_test_config_file)

class TestScheduler(unittest.TestCase):
    """
    This class tests the scheduler from yambnopy.

    1. Deploy a basic configuration file in ~/.yambopy
    compare the produced bash scripts to be run reference files
    """
    def test_default(self):
        """
        run the echo command on the scheduler using the default scheduler
        """
        s = Scheduler.factory(cores=1)
        s.add_module("abinit")
        s.add_command("echo 'hello'")
        print(header("default: %s"%s.__class__))
        print(s)
        s.write(_test_file)

        #remove files
        os.remove(_test_file)

    def test_print(self):
        """
        run the echo command on the different schedulers
        """
        #create basic configuration file
        init_config()

        #run using that configuration file
        for schedulername in ["oar","pbs","bash"]:
            print(header(schedulername))
            s = Scheduler.factory(scheduler=schedulername,cores=1,nodes=2)
            s.add_module("abinit")
            s.add_command("echo 'hello'")
            print(s)

        #remove files
        clean_config()

class TestSchedulerRun(unittest.TestCase):
    """
    This class tests the scheduler from yambnopy.

    2. Execute the echo command and check if the local configuration is working
    """
    def test_default(self):
        """
        run a echo hello world using the default scheduler on this machine
        """
        #run using that configuration file
        s = Scheduler.factory(cores=1)
        print(header("default: %s"%s.__class__))
        s.add_module("abinit")
        s.add_command("echo 'hello world'")
        s.run()

    def test_run(self):
        """
        run a echo hello world on the different possible configurations
        """
        init_config()

        for schedulername in ["oar","pbs","bash","slurm"]:
            print(header(schedulername))
            s = Scheduler.factory(scheduler=schedulername,cores=1)
            s._config = _test_config_file
            s.add_module("abinit")
            s.add_command("echo 'hello world'")
            s.run(dry=True)

        #remove files
        clean_config()


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Test the yambopy script.')
    parser.add_argument('-i', '--input', action="store_true",
                        help='Generate the bash files and compare with the reference ones')
    parser.add_argument('-f', '--full',  action="store_true",
                        help='Generate the bash files, run them and compare the results')
    parser.add_argument('-c', '--clean',  action="store_true",
                        help='Clean all the data from a previous run')
    args = parser.parse_args()

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    # Count the number of errors
    nerrors = 0

    ul = unittest.TestLoader()
    tr = unittest.TextTestRunner(verbosity=2)

    # Run the test
    if args.input:
        suite = ul.loadTestsFromTestCase(TestScheduler)
        nerrors += not tr.run(suite).wasSuccessful()

    if args.full:
        suite = ul.loadTestsFromTestCase(TestSchedulerRun)
        nerrors += not tr.run(suite).wasSuccessful()

    #clean tests
    if args.clean:
        print("cleaning...")
        os.system('rm -rf scf')
        print("done!")

    sys.exit(nerrors)



from qepy import  Path,ProjwfcXML,ProjwfcIn
import matplotlib.pyplot as plt

# k-points map
npoints = 50
path_kpoints = Path([ [[0.0, 0.0, 0.0 ],'G'],
                      [[0.0, 0.0, 1.0 ],'H'],
                      [[1./2,0.0,1./2.],'N'],
                      [[0.0, 0.0, 0.0 ],'G'],
                      [[1./2, 1./2, 1./2 ],'P'],
                      [[1./2,0.0,1./2. ],'N']], [npoints,npoints,npoints,npoints,npoints])

atom_s = [8]
atom_p = [0,1,2]
atom_d = [3,4,5,6,7]

# Class Projwfc
# Class to run projwfc.x and create 
# atomic_proj.xml (comment if already done)
'''
proj = ProjwfcIn(prefix='pw')
proj.run(folder='bands')
'''

# Class ProjwfcXML
# Atom-projected band structure. Colormap
band = ProjwfcXML(prefix='pw',path='bands')

# Manual selection of the lists of states by inspecting projwfc output
#s = [8]
#p = [0,1,2]
#d = [3,4,5,6,7]

# Automatic selection of the states
s = band.get_states_helper(orbital_query=['s'])
p = band.get_states_helper(orbital_query=['p'])
d = band.get_states_helper(orbital_query=['d'])

fig = plt.figure(figsize=(5,7))
ax  = fig.add_axes( [ 0.12, 0.10, 0.70, 0.80 ])

band.plot_eigen(ax,path_kpoints=path_kpoints,cmap='viridis',cmap2='rainbow',selected_orbitals=p,selected_orbitals_2=d)

# Plot colormap
#
import matplotlib as mpl
cmap =plt.get_cmap('viridis')
cmap2=plt.get_cmap('rainbow')
bx  = fig.add_axes( [ 0.84, 0.10, 0.03, 0.80 ])
cx  = fig.add_axes( [ 0.93, 0.10, 0.03, 0.80 ])
norm = mpl.colors.Normalize(vmin=0.,vmax=1.)
cb1 = mpl.colorbar.ColorbarBase(bx, cmap=cmap, norm=norm,orientation='vertical',ticks=[0,1])
cb1.set_ticklabels(['d', 'p'])
cb2 = mpl.colorbar.ColorbarBase(cx, cmap=cmap2, norm=norm,orientation='vertical',ticks=[0,1])
cb2.set_ticklabels(['d', 'p'])

plt.show()



##############################################################################
#
# Author: Alejandro Molina-Sanchez
# Run real-time simulations with yambo
# 
# Warning: Real-time simulations requires several data folders for running
# properly. Before using this scripts compulsively is recommended
# to understand the different run levels.
#
# This script breaks the symmetries of a given nscf run and
# prepares the rt-folder for the RT simulations
#
##############################################################################
from __future__ import print_function
from yambopy     import *
import argparse

print('This script breaks the symmetries of a given nscf run and')
print('prepares the rt-folder for the RT simulations')
print('Any previous "database" folder will be erased')
print('arg1: folder with nscf data')
print('arg2: folder for RT simulation')
print('arg3: prefix')
print('arg4: symmetry')
print('Example: python break-symm.py -i nscf -o rt-100 -p si -s 100')

parser = argparse.ArgumentParser(description='Map of a double-grid')
parser.add_argument('-i' ,'--input'    , help='Folder with nscf data')
parser.add_argument('-o' ,'--output'   , help='Folder for RT simulation')
parser.add_argument('-p' ,'--prefix'   ,help='Prefix of nscf calculation')
parser.add_argument('-s' ,'--symmetry' ,help='Choice symmetry: 100 010 110')
args = parser.parse_args()

print('Folder of nscf data     ===>>>  ' ,args.input)
print('Folder of RT simulation ===>>>  ' ,args.output)
print('Prefix                  ===>>>  ' ,args.prefix)
print('Symmetry                ===>>>  ' ,args.symmetry)

nscf_folder = args.input
rt_folder   = args.output
prefix      = args.prefix
symm        = args.symmetry
  
# Check rt_folder
if os.path.isdir(rt_folder):
    print('Real-time folder already exists')
    raise SystemExit
# Generation of the database folder
if os.path.isdir('database'):
    os.system('rm -rf database')
os.system('cd %s/%s.save ; p2y -O ../../database' % (nscf_folder, prefix))
os.system('cd database; yambo')
  
# Breaking of symmetries

if symm   == '100':
    breaking_symmetries([1,0,0], [0,0,0], rt_folder)
elif symm == '010':
    breaking_symmetries([0,1,0], [0,0,0], rt_folder)
elif symm == '110':
    breaking_symmetries([1,0,0], [0,1,0], rt_folder)



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC, AMS, FP, RR, MN
#
# This file is part of the yambopy project
#
import os
import numpy as np
import matplotlib.pyplot as plt
from netCDF4 import Dataset
from itertools import product
from yambopy.units import ha2ev, I
from yambopy.plot.plotting import add_fig_kwargs,BZ_Wigner_Seitz
from yambopy.lattice import replicate_red_kmesh, calculate_distances, car_red, red_car
from yambopy.kpoints import get_path, get_path_car
from yambopy.tools.funcs import gaussian, lorentzian, boltzman_f, abs2
from yambopy.tools.string import marquee
from yambopy.tools.types import CmplxType
from yambopy.plot.bandstructure import YambopyBandStructure
from yambopy.tools.skw import SkwInterpolator
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.dbs.electronsdb import YamboElectronsDB
from yambopy.dbs.qpdb import YamboQPDB
from yambopy.io.cubetools import write_cube
from yambopy.bse.realSpace_excitonwf import ex_wf2Real
from yambopy.bse.rotate_excitonwf import rotate_exc_wf

class ExcitonList():
    """
    Container class to perform operations on lists of excitons
    """
    def __init__(self,excitonlist):
        self.excitonlist = excitonlist

    def __str__(self):
        lines = []; app = lines.append
        for exciton in self.excitonlist:
            app( str(exciton.get_string(singleline=True)) )
        return "\n".join(lines)

class Exciton():
    """
    Basic container of data for a single exciton
    TODO: classify the excitons according to their symmetry
    """
    def __init__(self,energy,intensity,degeneracy,coeffs=None,wf=None):
        self.energy = energy
        self.intensity = intensity
        self.degeneracy = degeneracy
        self.coeffs = coeffs
        self.wf = wf

    def get_string(self,singleline=False):
        lines = []; app = lines.append
        app( 'energy:     %8.4'%self.energy )
        app( 'intensity:  %8.4'%self.intensity )
        app( 'degeneracy: %8d'%self.degeneracy )
        if singleline: return "".join(lines)
        return "\n".join(lines)
    
    def __str__(self):
        return self.get_string()

class YamboExcitonDB(object):
    """ Read the excitonic states database from yambo

        Exciton eigenvectors are arranged as eigenvectors[i_exc, i_kvc]
        Transitions are unpacked in table[ i_k, i_v, i_c, i_s_c, i_s_v ] (last two are spin indices)
    """
    def __init__(self,lattice,Qpt,eigenvalues,l_residual,r_residual,spin_pol='no',car_qpoint=None,q_cutoff=None,Lkind=None,table=None,eigenvectors=None):
        if not isinstance(lattice,YamboLatticeDB):
            raise ValueError('Invalid type for lattice argument. It must be YamboLatticeDB')

        self.Qpt = Qpt
        self.lattice = lattice
        self.eigenvalues = eigenvalues
        self.l_residual = l_residual
        self.r_residual = r_residual
        #optional
        self.car_qpoint = car_qpoint
        self.q_cutoff = q_cutoff
        self.table = table
        self.Lkind    = Lkind
        if table is not None:
            self.bs_bands = np.array([np.min(self.table[:,1]),np.max(self.table[:,2])]) # set range of bse bands
        self.eigenvectors = eigenvectors
        self.spin_pol = spin_pol

    @classmethod
    def from_db_file(cls,lattice,filename='ndb.BS_diago_Q1',folder='.',Load_WF=True, neigs=-1):
        """ 
        Initialize this class from a file

        Set `Read_WF=False` to avoid reading eigenvectors for faster IO and memory efficiency.
        If neigs < 0 ; all eigen values (vectors) are loaded or else first neigs are loaded 
        " In case of non-TDA, we load right eigenvectors.
        """
        path_filename = os.path.join(folder,filename)
        if not os.path.isfile(path_filename):
            raise FileNotFoundError("File %s not found in YamboExcitonDB"%path_filename)

        # Qpoint
        Qpt = filename.split("Q",1)[1]

        with Dataset(path_filename) as database:
            #energies
            eig =  database.variables['BS_Energies'][...].data*ha2ev
            is_coupling = database.variables['COUPLING'][...].data
            eigenvalues = eig[:,0]+eig[:,1]*I
            neig_full = len(eigenvalues)
            ## in the case of Coupling, we override partial loading. This is because, the energies are 
            ## not ordered, so it really does not make sense.
            if is_coupling and neigs >0:
                print("Warning : In the coupling case, all the states are read, overriding user input")
            #
            if neigs < 0 or neigs > neig_full or is_coupling:
                neigs = neig_full
            #
            eigenvalues = eigenvalues[:neigs]

            # Check Lkind if present
            Lkind = None
            if 'X_kind' in list(database.variables.keys()):
                Lkind = ''.join( database.variables['X_kind'][0].astype(str) ).strip()

            if 'BS_left_Residuals' in list(database.variables.keys()):
                # MN: using complex views instead of a+I*b copies to avoid memory duplication
                # Old (yet instructive) memory duplication code
                #rel,iml = database.variables['BS_left_Residuals'][:].T
                #rer,imr = database.variables['BS_right_Residuals'][:].T
                #l_residual = rel+iml*I
                #r_residual = rer+imr*I
                l_residual = database['BS_left_Residuals'][:neigs,...].data
                r_residual = database['BS_right_Residuals'][:neigs,...].data
                l_residual = l_residual.view(dtype=CmplxType(l_residual)).reshape(len(l_residual))
                r_residual = r_residual.view(dtype=CmplxType(r_residual)).reshape(len(r_residual))
            if 'BS_Residuals' in list(database.variables.keys()):
                # Compatibility with older Yambo versions
                rel,iml,rer,imr = database['BS_Residuals'][:neigs,...].data.T
                l_residual = rel+iml*I
                r_residual = rer+imr*I
            if 'BS_L_magn_Residuals' in list(database.variables.keys()):
                #residuals
                rel,iml = database.variables['BS_L_magn_Residuals'][:].T
                rer,imr = database.variables['BS_L_magn_Residuals'][:].T
                l_residual = rel+iml*I
                r_residual = rer+imr*I

            car_qpoint = None
            # Finite momentum
            if 'BS_Q' in list(database.variables.keys()):                         # Lumen compatibility
                car_qpoint = database.variables['BS_Q'][...].data/lattice.alat
            elif 'Q-point' in list(database.variables.keys()):
                car_qpoint = database.variables['Q-point'][...].data/lattice.alat # Yambo compatibility
            if Qpt=="1": car_qpoint = np.zeros(3)

            #eigenvectors
            table = None
            eigenvectors = None
            if Load_WF and 'BS_EIGENSTATES' in database.variables:
                eiv = database['BS_EIGENSTATES'][:neigs,...].data
                #eiv = eiv[:,:,0] + eiv[:,:,1]*I
                #eigenvectors = eiv
                eigenvectors = eiv.view(dtype=CmplxType(eiv)).reshape(eiv.shape[:-1])

            table = np.rint(database.variables['BS_TABLE'][:].T).astype(int)
            spin_vars = [int(database.variables['SPIN_VARS'][:][0]), int(database.variables['SPIN_VARS'][:][1])]
            if spin_vars[0] == 2 and spin_vars[1] == 1:
               spin_pol = 'pol'
            else:
               spin_pol = 'no'
        # Check if Coulomb cutoff is present
        path_cutoff = os.path.join(path_filename.split('ndb',1)[0],'ndb.cutoff')  
        q_cutoff = None
        if os.path.isfile(path_cutoff):
            with Dataset(path_cutoff) as database:
                bare_qpg = database.variables['CUT_BARE_QPG'][:]
                bare_qpg = bare_qpg[:,:,0]+bare_qpg[:,:,1]*I
                q_cutoff = np.abs(bare_qpg[0,int(Qpt)-1])

        return cls(lattice,Qpt,eigenvalues,l_residual,r_residual,spin_pol,q_cutoff=q_cutoff,car_qpoint=car_qpoint,Lkind=Lkind,table=table,eigenvectors=eigenvectors)

    @property
    def unique_vbands(self):
        return np.unique(self.table[:,1]-1)

    @property
    def unique_cbands(self):
        return np.unique(self.table[:,2]-1)

    @property
    def transitions_v_to_c(self):
        """Compute transitions from valence to conduction"""
        if hasattr(self,"_transitions_v_to_c"): return self._transitions_v_to_c
        uniq_v = self.unique_vbands
        uniq_c = self.unique_cbands
        transitions_v_to_c = dict([ ((v,c),[]) for v,c in product(uniq_v,uniq_c) ])

        #add elements to dictionary
        kidx = set()
        for eh,kvc in enumerate(self.table-1):
            k,v,c = kvc[0:3]
            kidx.add(k)
            transitions_v_to_c[(v,c)].append((k,eh))
        self.nkpoints = len(kidx)

        #make an array 
        for t,v in list(transitions_v_to_c.items()):
            if len(np.array(v)):
                transitions_v_to_c[t] = np.array(v)
            else:
                del transitions_v_to_c[t]

        self._transitions_v_to_c = transitions_v_to_c 
        return transitions_v_to_c

    @property
    def nkpoints(self): return max(self.table[:,0])

    @property
    def nvbands(self): return len(self.unique_vbands)

    @property
    def ncbands(self): return len(self.unique_cbands)

    @property
    def nbands(self): return self.ncbands+self.nvbands

    @property
    def mband(self): return max(self.unique_cbands)+1
 
    @property
    def ntransitions(self): return len(self.table)

    @property
    def nexcitons(self): return len(self.eigenvalues)
    
    @property
    def start_band(self): return min(self.unique_vbands)

    def write_sorted(self,prefix='yambo'):
        """
        Write the sorted energies and intensities to a file
        """
        #get intensities
        eig = self.eigenvalues.real
        intensities = self.get_intensities()

        #get sorted energies
        sort_e, sort_i = self.get_sorted()

        #write excitons sorted by energy
        se_arr = np.array(sort_e)
        n_idx = se_arr[:, 1].astype(int)
        data_e = np.column_stack((se_arr[:, 0], intensities[n_idx], n_idx + 1))
        np.savetxt('%s_E.dat'%prefix, data_e, fmt='%16.8f %20.8e %10d',
                   header='    E [ev]             Strength           Index')

        #write excitons sorted by intensities
        si_arr = np.array(sort_i)
        n_idx = si_arr[:, 1].astype(int)
        data_i = np.column_stack((eig[n_idx], np.abs(si_arr[:, 0]), n_idx + 1))
        np.savetxt('%s_I.dat'%prefix, data_i, fmt='%16.8f %20.8e %10d',
                   header='    E [ev]             Strength           Index')

    def get_Akcv(self):
        """
        Convert eigenvectors from (neigs,BS_table) -> (neigs,nblks,nspin,k,c,v)
        nblks = 2 for coupling, else 1 for TDA
        """
        nspin = 1
        if self.spin_pol == 'pol': nspin = 2
        #
        tmp_akcv = getattr(self, 'Akcv', None) 
        if tmp_akcv is not None: return tmp_akcv 
        #
        if self.eigenvectors is None: return None
        eig_wfcs = self.eigenvectors
        #
        nk = self.nkpoints
        nv = self.nvbands
        nc = self.ncbands
        # Make sure nspin * nc * nv * nk = BS_TABLE length
        table_len = nspin*nk*nv*nc
        assert table_len == self.table.shape[0], "BS_TABLE length not equal to ns * nc * nv * nk"
        #
        v_min = np.min(self.table[:,1])
        c_min = np.min(self.table[:,2])
        bs_table0 = self.table[:,0]-1
        bs_table1 = self.table[:,1] - v_min
        bs_table2 = self.table[:,2] - c_min
        bs_table3 = self.table[:,3]-1
        #
        eig_wfcs_returned = np.zeros(eig_wfcs.shape,dtype=eig_wfcs.dtype)
        #
        sort_idx = bs_table0*nc*nv + bs_table2*nv + bs_table1 + nk*nc*nv*bs_table3
        #
        eig_wfcs_returned[:,sort_idx] = eig_wfcs[...,:table_len]
        # check if this is coupling .
        if eig_wfcs.shape[-1]//table_len == 2:
            eig_wfcs_returned[:,sort_idx+table_len] = eig_wfcs[...,table_len:]
            # NM : Note that here v and c are inverted i.e 
            # psi_S = Akcv * phi_v(r_e) * phi_c^*(r_h)
            eig_wfcs_returned = eig_wfcs_returned.reshape(-1,2,nspin,nk,nc,nv)
        else :
            eig_wfcs_returned = eig_wfcs_returned.reshape(-1,1,nspin,nk,nc,nv)
        #
        self.Akcv = eig_wfcs_returned
        return self.Akcv
    
    def real_wf_to_cube(self, iexe, wfdb, fixed_position=[0,0,0], supercell=[1,1,1], degen_tol=1e-2,
                        wfcCutoffRy=-1, fix_particle='h', phase=False, block_size=256):
        """
        Function to compute and save real-space exciton wavefunctions and 
        dump to cube file

        Args:
            iexe: index of excitonic states (python indexing. so starts with 0)
            wfcb: wavefunction database.
            fixed_position (list): Position of fixed particle in crystal coordinates
            supercell (list): Supercell dimensions [nx,ny,nz]
            degen_tol (float): degeneracy threshold (in eV). default 0.01 eV
            fix_particle (str): 'e' to fix electron, 'h' to fix hole (default)
            wfcCutoffRy (float): Wavefunction cutoff in Rydberg (-1 for full cutoff)
            phase (bool): If True, include phase information i.e multiply the density with
                     sign of real part of the wavefunction
            block_size (int): Block size for memory-efficient computation. leave it to default
                        unless you are in exteremely low memory situation.

        Returns:
            None (write cube file to disk)
        """
        #
        # first get all degenerate states
        iexe_degen_states = np.array(self.get_degenerate(iexe+1,eps=degen_tol))-1
        print("Degenerate states: ",iexe_degen_states+1)
        # nicely arrange eigvectors to Akcv
        Akcv = self.get_Akcv()[iexe_degen_states]
        excQpt = self.car_qpoint
        # Convert the q-point to crystal coordinates
        Qpt = wfdb.ydb.lat @ excQpt
        #
        if fix_particle == 'h': name_file = 'electron'
        else: name_file = 'hole'

        if phase and wfdb.wf.shape[1] != 1 and wfdb.wf.shape[3] != 1:
            print("phase plot only works for nspin = 1 and nspinor == 1")
            phase = False
        if phase and len(iexe_degen_states) > 1:
            phase = False
            print("Warning: phase plots do not work for degenerate states")

        print('Computing exciton wavefunction (%s density) to real space.' %(name_file))
        sc_latvecs, atom_nums, atom_pos, real_wfc = ex_wf2Real(Akcv, Qpt, wfdb, [np.min(self.table[:, 1]),
                                                    np.max(self.table[:, 2])], fixed_position=fixed_position,
                                                    fix_particle=fix_particle, supercell=supercell,
                                                    wfcCutoffRy=wfcCutoffRy, block_size=block_size)
        # Compute the absoulte^2
        density = np.abs(real_wfc)**2
        # Multiply with phase if necessary
        if phase:
            phase = np.sign(real_wfc.real) #np.sign(np.angle(real_wfc))
            density *= phase
        #
        # sum over spinor indices and degenerate states
        real_wfc = np.sum(density,axis=(0,1,2,3))
        # normalize with max value
        max_normalize_val = np.max(np.abs(real_wfc))
        print('Max Normalization value: ',max_normalize_val)
        real_wfc *= (1.0/max_normalize_val)
        # write to cube file 
        print('Writing to .cube file')
        write_cube('exe_wf_%s_Qpt%s_state%d.cube' %(name_file,self.Qpt,iexe+1),
                   real_wfc, sc_latvecs, atom_pos, atom_nums,
                   header='Real space exciton wavefunction')

    def total_crys_angular_momentum(self, wfdb, iexe, symm_mat_cat, frac_vec_cart, degen_tol=1e-3, Dmats=None):
        """
        Computes the total crystal angular momentum along a given rotational symmetry axis.

        This method calculates the simultaneous eigenbasis of the Hamiltonian and the
        Symmetry operator. It verifies if the symmetry belongs to the little group of q
        before proceeding with the representation calculations.

        Parameters
        ----------
        wfdb : object
            The wavefunction database object containing lattice and basis information.
        iexe : int
            Index of the exciton (0-based Python indexing).
        symm_mat_cat : array_like of shape (3, 3)
            Rotational symmetry matrix in Cartesian coordinates.
        frac_vec_cart : array_like of shape (3,)
            Fractional translational vector associated with the symmetry operation.
        degen_tol : float, optional
            Tolerance for identifying degenerate excitons in eV. Default is 1e-3.
        Dmats : array_like, optional
            Representation matrices for the given symmetry. If None, they are computed
            directly from the ``wfdb``. Default is None.

        Returns
        -------
        list
            A list containing:
            - w : ndarray
                The real part of the angular momentum values (eigenvalues).
            - sbasis_r : ndarray
                The right symmetrized basis vectors.
            - sbasis_l : ndarray, optional
                The left symmetrized basis vectors. This is included only if the
                exciton wavefunction has distinct left/right components (non-Hermitian).

        Returns None
        ------------
        None
            Returned if the rotation is improper (determinant < 0) or if the
            symmetry does not belong to the little group of Q.
        """
        # Identify degenerate excitons
        iexe = np.array(self.get_degenerate(iexe + 1, eps=degen_tol), dtype=int) - 1
        print(f"Number of degenerate excitons found : {len(iexe)}. List : {iexe}")

        # Check for improper rotation
        det_r = np.linalg.det(symm_mat_cat)
        assert det_r > 0, "Warning : Improper rotation, Only rotation are allowed."

        # Prepare lattice vectors and transformation matrices
        lat_vec = wfdb.ydb.lat
        lat_vec_inv = np.linalg.inv(lat_vec)
        symm_mat_red = lat_vec @ symm_mat_cat @ lat_vec_inv

        # Convert the q-point to crystal coordinates
        exc_qpt_car = self.car_qpoint
        print(exc_qpt_car)
        exc_qpt_crys = lat_vec @ exc_qpt_car
        print(f"Qpt : ({exc_qpt_crys[0]:.6f}, {exc_qpt_crys[1]:.6f}, {exc_qpt_crys[2]:.6f})")

        # Verify if symmetry belongs to the little group of Q
        sq_minus_q = np.einsum('ij,j->i', symm_mat_red, exc_qpt_crys) - exc_qpt_crys
        sq_minus_q = sq_minus_q - np.rint(sq_minus_q)

        assert np.linalg.norm(sq_minus_q) < 1e-4, "Warning : The given symmetry does not belong to little group of Q."

        # Compute representation matrices if not provided
        if Dmats is None:
            Dmats = wfdb.Dmat(
                symm_mat=symm_mat_cat.reshape(1, 3, 3),
                frac_vec=frac_vec_cart.reshape(1, 3),
                time_rev=False
            )[0]

        # Retrieve exciton wavefunctions
        # Akcv represents the coefficients of the exciton wavefunction
        akcv = self.get_Akcv()
        akcv_left = akcv

        # Handle non-Hermitian cases where left eigenvectors differ
        if akcv.shape[1] == 2:
            print("Computing left ev ...")
            akcv_left = np.linalg.inv(akcv.reshape(len(akcv), -1)).conj().T.reshape(akcv.shape)

        # Select specific degenerate components
        ak_r = akcv[iexe]
        ak_l = akcv_left[iexe].conj()

        # Calculate phase factor
        tau_dot_k = np.exp(1j * 2 * np.pi * np.dot(self.car_qpoint, frac_vec_cart))

        # Rotate the right wavefunction
        rot_akcv = rotate_exc_wf(
            ak_r, symm_mat_red, wfdb.kBZ, exc_qpt_crys,
            Dmats, False, ktree=wfdb.ktree
        )

        # Compute the representation matrix
        rep = tau_dot_k * np.einsum('m...,n...->mn', ak_l, rot_akcv, optimize=True)
        w, v = np.linalg.eig(rep)

        # Calculate rotation angle and axis via SVD
        _, _, vt = np.linalg.svd(symm_mat_cat - np.eye(3))
        # axis = vt[-1] / np.linalg.norm(vt[-1]) # axis is calculated but not currently used in return
        angle = np.arccos(np.clip((np.trace(symm_mat_cat) - 1) / 2, -1.0, 1.0))

        # Convert eigenvalues to angular momentum
        w = 1j * np.log(w) / angle
        w = w.real

        # Transform basis vectors
        sbasis_r = np.einsum('ij,i...->j...', v, ak_r, optimize=True)

        if akcv.shape[1] == 2:
            sbasis_l = np.einsum('ij,i...->j...', v, ak_l.conj(), optimize=True)
            return [w, sbasis_r, sbasis_l]
        else:
            return [w, sbasis_r]


    def get_nondegenerate(self,eps=1e-4):
        """
        get a list of non-degenerate excitons
        """
        non_deg_e   = [0]
        non_deg_idx = [] 

        #iterate over the energies
        for n,e in enumerate(self.eigenvalues):
            if not np.isclose(e,non_deg_e[-1],atol=eps):
                non_deg_e.append(e)
                non_deg_idx.append(n)

        return np.array(non_deg_e[1:]), np.array(non_deg_idx)

    def get_intensities(self):
        """
        get the intensities of the excitons
        """
        intensities = np.abs(self.l_residual*self.r_residual)
        intensities /= np.max(intensities)
        return intensities

    def get_sorted(self):
        """
        Return the excitonic weights sorted according to energy and intensity
        """
        #get intensities
        eig = self.eigenvalues.real
        intensities = self.get_intensities()

        #list ordered with energy
        sort_e = sorted(zip(eig, list(range(self.nexcitons))))

        #list ordered with intensity
        sort_i = sorted(zip(intensities, list(range(self.nexcitons))),reverse=True)

        return sort_e, sort_i 

    def get_degenerate(self,index,eps=1e-4,rtol=1e-3):
        """
        Get degenerate excitons
        
        Args:
            eps: maximum energy difference to consider the two excitons degenerate in eV
        """
        energy = self.eigenvalues[index-1].real
        excitons = np.where(np.isclose(self.eigenvalues.real, energy, atol=eps, rtol=rtol))[0] + 1
        return excitons.tolist()

    def exciton_bs(self,energies,path,excitons=(0,),debug=False):
        """
        Calculate exciton band-structure
            
            Arguments:
            energies -> can be an instance of YamboElectronsDB or YamboQPDB
            path     -> Path object in reduced coordinates to use for plotting the band structure
            exciton  -> exciton index to plot
            spin     -> So far only spin_pol='no' (or spin-up only) is implemented. Can be extended to spin_pol='pol' (spin-polarized)
        """
        if self.eigenvectors is None:
            raise ValueError('This database does not contain Excitonic states,'
                              'please re-run the yambo BSE calculation with the WRbsWF option in the input file.')
        if isinstance(excitons, int):
            excitons = (excitons,)

        car_kpoints = self.lattice.car_kpoints
        rlat        = self.lattice.rlat
        bands_kpoints, band_indexes, path_car = get_path(car_kpoints,rlat,None,path,debug=debug) # None means the kpoints are already expanded

        if debug:
            import matplotlib.pyplot as plt
            kpoints = self.lattice.red_kpoints
            rep = list(range(-1,2))
            kpoints_rep, kpoints_idx_rep = replicate_red_kmesh(kpoints,repx=rep,repy=rep,repz=rep)
            for i,k in zip(band_indexes,bands_kpoints):
                x,y,z = k
                plt.text(x,y,i) 
            plt.scatter(kpoints_rep[:,0],kpoints_rep[:,1])
            plt.plot(path.kpoints[:,0],path.kpoints[:,1],c='r')
            plt.scatter(bands_kpoints[:,0],bands_kpoints[:,1])
            plt.show()
            exit()

        #get eigenvalues along the path
        max_val = max(self.unique_vbands)
        if isinstance(energies,YamboElectronsDB):
            #expand eigenvalues to the full brillouin zone
            if not energies.EXPAND: energies.expandEigenvalues()
            exc_energies = energies.eigenvalues[0] # SPIN-UP CHANNEL ONLY      
        elif isinstance(energies,YamboQPDB):
            if energies.min_band-1>self.start_band or energies.max_band<self.mband:
                print("[WARNING] BSE band range: {self.start_band+1} - {self.mband}\n \
                                 QP  band range: {energies.min_band} - {energies.max_band}")
            #expand the quasiparticle energies to the full brillouin zone
            QP_energies = energies.expand_eigenvalues(self.lattice)
            #deal with index mismatches
            max_band  =max(energies.max_band,self.mband)
            exc_energies = np.zeros((len(QP_energies),max_band))
            exc_energies[:,energies.min_band-1:energies.max_band] = QP_energies
            max_val = max_val - energies.min_band+1
        else:
            raise ValueError("Energies argument must be an instance of YamboElectronsDB or YamboQPDB. Got %s"%(type(energies)))

        exc_weights = self.get_exciton_weights(excitons)      
        exc_energies = exc_energies[band_indexes]
        exc_weights  = exc_weights[band_indexes]

        #make top valence band to be zero
        exc_energies -= max(exc_energies[:,max_val])
        
        return bands_kpoints, exc_energies, exc_weights, path_car 

    def get_exciton_weights(self,excitons):
        """get weight of state in each band"""
        weights = np.zeros([self.nkpoints,self.mband])
        for exciton in excitons:
            #get the eigenstate
            eivec = self.eigenvectors[exciton-1]

            #add weights
            sum_weights = 0
            for t,kcv in enumerate(self.table):
                k,c,v = kcv[0:3]-1
                this_weight = abs2(eivec[t])
                weights[k,c] += this_weight
                weights[k,v] += this_weight
                sum_weights += this_weight
            if abs(sum_weights - 1) > 1e-3: raise ValueError('Excitonic weights does not sum to 1 but to %lf.'%sum_weights)

        return weights
    
    def get_exciton_total_weights(self,excitons):
        """get weight of state in each band"""
        total_weights = np.zeros(self.nkpoints)
        for exciton in excitons:
            #get the eigenstate
            eivec = self.eigenvectors[exciton-1]
            #add weights
            sum_weights = 0
            for t,kcv in enumerate(self.table):
#                k,c,v = kcv[0:3]
                k = kcv[0]
                total_weights[k-1] += abs2(eivec[t])
            if abs(sum(total_weights) - 1) > 1e-3: raise ValueError('Excitonic weights does not sum to 1 but to %lf.'%sum_weights)
 
        return total_weights

    def get_exciton_transitions(self,excitons):
        """get weight of state in each band"""
        # Double check the part of the array w_k_v_to_c
        # We should comment more this part
        #weights = np.zeros([self.nkpoints,self.mband])
        w_k_v_to_c = np.zeros([self.nkpoints,self.nvbands,self.ncbands])
        v_min = self.unique_vbands[0]
        c_min = self.unique_cbands[0]
        for exciton in excitons:
            #get the eigenstate
            eivec = self.eigenvectors[exciton-1]
            #add weights
            #sum_weights = 0
            for t,kcv in enumerate(self.table):
                k,c,v = kcv-1                             
                #k,v,c = kcv-1                 # bug?? Double-check
                this_weight = abs2(eivec[t])
                w_k_v_to_c[k,v-v_min,c-c_min] = this_weight   # new
            #if abs(sum_weights - 1) > 1e-3: raise ValueError('Excitonic weights does not sum to 1 but to %lf.'%sum_weights)
 
        #return weights, w_k_v_to_c
        return w_k_v_to_c

    def get_exciton_2D(self,excitons,f=None):
        """get data of the exciton in 2D"""
        weights = self.get_exciton_weights(excitons)
        #sum all the bands
        weights_bz_sum = np.sum(weights,axis=1)
        if f: weights_bz_sum = f(weights_bz_sum)

        kmesh_full, kmesh_idx = replicate_red_kmesh(self.lattice.red_kpoints,repx=range(-1,2),repy=range(-1,2))
        x,y = red_car(kmesh_full,self.lattice.rlat)[:,:2].T
        weights_bz_sum = weights_bz_sum[kmesh_idx]
        return x,y,weights_bz_sum

    def get_exciton_2D_spin_pol(self,excitons,f=None):
        """get data of the exciton in 2D for spin polarized calculations"""
        weights_up, weights_dw = self.get_exciton_weights_spin_pol(excitons)

        #sum all the bands
        weights_bz_sum_up = np.sum(weights_up,axis=1)
        weights_bz_sum_dw = np.sum(weights_dw,axis=1)

        if f: weights_bz_sum_up = f(weights_bz_sum_up)
        if f: weights_bz_sum_dw = f(weights_bz_sum_dw)

        kmesh_full, kmesh_idx = replicate_red_kmesh(self.lattice.red_kpoints,repx=range(-1,2),repy=range(-1,2))
        x,y = red_car(kmesh_full,self.lattice.rlat)[:,:2].T
        weights_bz_sum_up = weights_bz_sum_up[kmesh_idx]
        weights_bz_sum_dw = weights_bz_sum_dw[kmesh_idx]

        return x,y,weights_bz_sum_up,weights_bz_sum_dw
 
 
    def plot_exciton_2D_ax(self,ax,excitons,f=None,mode='hexagon',limfactor=0.8,spin_pol=None,**kwargs):
        """
        Plot the exciton weights in a 2D Brillouin zone
       
           Arguments:
            excitons -> list of exciton indexes to plot
            f -> function to apply to the exciton weights. Ex. f=log will compute the 
                 log of th weight to enhance the small contributions
            mode -> possible values are 'hexagon'/'square' to use hexagons/squares as markers for the 
                    weights plot and 'rbf' to interpolate the weights using radial basis functions.
            limfactor -> factor of the lattice parameter to choose the limits of the plot 
            scale -> size of the markers
        """
        if spin_pol is not None: print('Plotting exciton mad in 2D axis for spin polarization: %s' % spin_pol)

        if spin_pol is not None:
           x,y,weights_bz_sum_up,weights_bz_sum_dw = self.get_exciton_2D_spin_pol(excitons,f=f)
        else:
           x,y,weights_bz_sum = self.get_exciton_2D(excitons,f=f)

        weights_bz_sum=weights_bz_sum/np.max(weights_bz_sum)
        
        #filter points outside of area
        lim = np.max(self.lattice.rlat)*limfactor
        dlim = lim*1.1
        if spin_pol is not None:
           filtered_weights_up = [[xi,yi,di] for xi,yi,di in zip(x,y,weights_bz_sum_up) if -dlim<xi and xi<dlim and -dlim<yi and yi<dlim]
           filtered_weights_dw = [[xi,yi,di] for xi,yi,di in zip(x,y,weights_bz_sum_dw) if -dlim<xi and xi<dlim and -dlim<yi and yi<dlim]
           x,y,weights_bz_sum_up = np.array(filtered_weights_up).T
           x,y,weights_bz_sum_dw = np.array(filtered_weights_dw).T
        else:
           filtered_weights = [[xi,yi,di] for xi,yi,di in zip(x,y,weights_bz_sum) if -dlim<xi and xi<dlim and -dlim<yi and yi<dlim]
           x,y,weights_bz_sum = np.array(filtered_weights).T
        # Add contours of BZ
        ax.add_patch(BZ_Wigner_Seitz(self.lattice))

        #plotting
        if mode == 'hexagon': 
            scale = kwargs.pop('scale',1)
            if spin_pol=='up':
               s=ax.scatter(x,y,s=scale,marker='H',c=weights_bz_sum_up,rasterized=True,**kwargs)
            elif spin_pol=='dw':
               s=ax.scatter(x,y,s=scale,marker='H',c=weights_bz_sum_dw,rasterized=True,**kwargs)
            else:
               s=ax.scatter(x,y,s=scale,marker='H',c=weights_bz_sum,rasterized=True,**kwargs)
            ax.set_xlim(-lim,lim)
            ax.set_ylim(-lim,lim)
        elif mode == 'square': 
            scale = kwargs.pop('scale',1)
            if spin_pol=='up':
               s=ax.scatter(x,y,s=scale,marker='s',c=weights_bz_sum_up,rasterized=True,**kwargs)
            elif spin_pol=='dw':
               s=ax.scatter(x,y,s=scale,marker='s',c=weights_bz_sum_dw,rasterized=True,**kwargs)
            else:
               s=ax.scatter(x,y,s=scale,marker='s',c=weights_bz_sum,rasterized=True,**kwargs)
            ax.set_xlim(-lim,lim)
            ax.set_ylim(-lim,lim)
        elif mode == 'rbf':
            from scipy.interpolate import Rbf
            npts = kwargs.pop('npts',100)
            interp_method = kwargs.pop('interp_method','bicubic')
            if spin_pol=='up':
               rbfi = Rbf(x,y,weights_bz_sum_up,function='linear')
               x = y = np.linspace(-lim,lim,npts)
               weights_bz_sum_up = np.zeros([npts,npts])
            elif spin_pol=='dw':
               rbfi = Rbf(x,y,weights_bz_sum_dw,function='linear')
               x = y = np.linspace(-lim,lim,npts)
               weights_bz_sum_dw = np.zeros([npts,npts])
            else:
               rbfi = Rbf(x,y,weights_bz_sum,function='linear')
               x = y = np.linspace(-lim,lim,npts)
               weights_bz_sum = np.zeros([npts,npts])

            for col in range(npts):
                if spin_pol=='up':
                   weights_bz_sum_up[:,col] = rbfi(x,np.ones_like(x)*y[col])
                elif spin_pol=='dw':
                   weights_bz_sum_dw[:,col] = rbfi(x,np.ones_like(x)*y[col])
                else:
                   weights_bz_sum[:,col] = rbfi(x,np.ones_like(x)*y[col])
            # NB we have to take the transpose of the imshow data to get the correct plot
            if spin_pol=='up':
               s=ax.imshow(weights_bz_sum_up.T,interpolation=interp_method,extent=[-lim,lim,-lim,lim])
            elif spin_pol=='dw':
               s=ax.imshow(weights_bz_sum_dw.T,interpolation=interp_method,extent=[-lim,lim,-lim,lim])
            else:
               s=ax.imshow(weights_bz_sum.T,interpolation=interp_method,extent=[-lim,lim,-lim,lim])
        title = kwargs.pop('title',str(excitons))

        ax.set_title(title)
        ax.set_aspect('equal')
        ax.set_xticks([])
        ax.set_yticks([])
       
        return ax,s

    def get_exciton_3D(self,excitons,f=None):
        """get data of the exciton in 2D"""
        weights = self.get_exciton_weights(excitons)
        #sum all the bands
        weights_bz_sum = np.sum(weights,axis=1)
        if f: weights_bz_sum = f(weights_bz_sum)

        kmesh_full, kmesh_idx = replicate_red_kmesh(self.lattice.red_kpoints,repx=range(-1,2),repy=range(-1,2))
        x,y,z = red_car(kmesh_full,self.lattice.rlat)[:,:3].T
        weights_bz_sum = weights_bz_sum[kmesh_idx]
        return x,y,z,weights_bz_sum
 
    def plot_exciton_3D_ax(self,ax,excitons,f=None,mode='hexagon',limfactor=0.8,**kwargs):
        """
        Plot the exciton weights in a 3D Brillouin zone
       
           Arguments:
            excitons -> list of exciton indexes to plot
            f -> function to apply to the exciton weights. Ex. f=log will compute the 
                 log of th weight to enhance the small contributions
            mode -> possible values are 'hexagon' to use hexagons as markers for the 
                    weights plot and 'rbf' to interpolate the weights using radial basis functions.
            limfactor -> factor of the lattice parameter to choose the limits of the plot 
            scale -> size of the markers
        """
        x,y,z,weights_bz_sum = self.get_exciton_3D(excitons,f=f)
        print(x,y,z,weights_bz_sum)
        exit()

        #filter points outside of area
        lim = np.max(self.lattice.rlat)*limfactor
        dlim = lim*1.1
        filtered_weights = [[xi,yi,di] for xi,yi,di in zip(x,y,weights_bz_sum) if -dlim<xi and xi<dlim and -dlim<yi and yi<dlim]
        x,y,z,weights_bz_sum = np.array(filtered_weights).T

        #plotting
        if mode == 'hexagon': 
            scale = kwargs.pop('scale',1)
            ax.scatter(x,y,s=scale,marker='H',c=weights_bz_sum,rasterized=True,**kwargs)
            ax.set_xlim(-lim,lim)
            ax.set_ylim(-lim,lim)
        elif mode == 'square': 
            scale = kwargs.pop('scale',1)
            ax.scatter(x,y,s=scale,marker='s',c=weights_bz_sum,rasterized=True,**kwargs)
            ax.set_xlim(-lim,lim)
            ax.set_ylim(-lim,lim)
        elif mode == 'rbf':
            from scipy.interpolate import Rbf
            npts = kwargs.pop('npts',100)
            interp_method = kwargs.pop('interp_method','bicubic')
            rbfi = Rbf(x,y,weights_bz_sum,function='linear')
            x = y = np.linspace(-lim,lim,npts)
            weights_bz_sum = np.zeros([npts,npts])
            for col in range(npts):
                weights_bz_sum[:,col] = rbfi(x,np.ones_like(x)*y[col])
            # NB we have to take the transpose of the imshow data to get the correct plot
            ax.imshow(weights_bz_sum.T,interpolation=interp_method,extent=[-lim,lim,-lim,lim])
        title = kwargs.pop('title',str(excitons))
        
        ax.set_title(title)
        ax.set_aspect('equal')
        ax.set_xticks([])
        ax.set_yticks([])
        ax.add_patch(BZ_Wigner_Seitz(self.lattice))
        return ax

    def plot_nbrightest_2D(self,emin=0,emax=10,estep=0.001,broad=0.1,
                           mode='rbf',scale=3,nrows=2,ncols=2,eps=1e-5):
        """
        Create a plot with chi and vertical bars for the brightest excitons
        Also plot the 2D wavefunctions of the brightest excitons.

          Arguments:
            emin,emax -> minimum and maximum energy range to plot chi
            estep -> energy step to plot chi
            broad -> broadening of the exciton peaks
            mode -> possible values are 'hexagon' to use hexagons as markers for the 
                    weights plot and 'rbf' to interpolate the weights using radial basis functions.
            scale -> size of the markers
            nrows,ncols -> number of rows and colums for the 2D plots (default: 2x2)
            eps -> threshold to find degenerate states
        """
        import matplotlib.pyplot as plt
        figexc = plt.figure()
        n_brightest = nrows*ncols
        figchi = self.plot_chi(emin=emin,emax=emax,estep=estep,broad=broad,n_brightest=n_brightest,show=False)
        #plot vertical bar on the brightest excitons
        exc_e,exc_i = self.get_sorted()
        sorted_exc = sorted(exc_i[:n_brightest],key = lambda x: x[1])
        for n,(i,idx) in enumerate(sorted_exc):
            ax = figexc.add_subplot(nrows,ncols,n+1)
            excitons = self.get_degenerate(idx,eps)
            self.plot_exciton_2D_ax(ax,excitons,scale=scale,mode=mode)
        return figchi,figexc

    def get_exciton_bs(self,energies_db,path,excitons,size=1,f=None,debug=False):
        """
        Get a YambopyBandstructure object with the exciton band-structure
        
            Arguments:
            ax          -> axis extance of matplotlib to add the plot to
            lattice     -> Lattice database
            energies_db -> Energies database, can be either ElectronsDB or QPDB
            path        -> Path in the brillouin zone
        """
        from qepy.lattice import Path
        if not isinstance(path,Path): 
            raise ValueError('Path argument must be a instance of Path. Got %s instead'%type(path))
    
        if self.spin_pol=='no':
            bands_kpoints, exc_energies, exc_weights, path_car = self.exciton_bs(energies_db, path, excitons, debug)
            exc_energies = exc_energies[:,self.start_band:self.mband]
            exc_weights  = exc_weights[:,self.start_band:self.mband]
        elif self.spin_pol=='pol':
            bands_kpoints, exc_energies, exc_weights, path_car = self.exciton_bs(energies_db, path, excitons, debug)
            exc_energies = exc_energies[:,self.start_band:self.mband]
            exc_weights  = exc_weights[:,self.start_band:self.mband]

        if f: exc_weights = f(exc_weights)
        size *= 1.0/np.max(exc_weights)
        ybs = YambopyBandStructure(exc_energies, bands_kpoints, weights=exc_weights, kpath=path_car, size=size)
        return ybs

    def plot_exciton_bs_ax(self,ax,energies_db,path,excitons,size=1,space='bands',f=None,debug=None):
        ybs = self.get_exciton_bs(energies_db,path,excitons,size=size,space=space,f=f,debug=debug)
        return ybs.plot_ax(ax)

    @add_fig_kwargs

    def plot_exciton_bs(self,energies_db,path,excitons,size=1,space='bands',f=None,debug=False,**kwargs):
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(1,1,1)
        self.plot_exciton_bs_ax(ax,energies_db,path,excitons,size=size,space=space,f=f,debug=debug)
        return fig

    def interpolate(self,energies,path,excitons,lpratio=5,f=None,verbose=1,size=1,**kwargs):
        """ 
        Interpolate exciton bandstructure using SKW interpolation from Abipy

        FP: the QPDB part needs to be tested
        """

        lat = self.lattice
        
        # These are the weights to be interpolated on top of the band structure
        weights = self.get_exciton_weights(excitons)
        weights = weights[:,self.start_band:self.mband]
        if f: weights = f(weights)
        size *= 1.0/np.max(weights)

        # Now we treat the band structure, which can be:
        ## - ElectronsDB -> DFT values
        ## - QPDB -> GW values
        ## - IBZ or BZ -> with or without symmetries
        use_symmetries = (isinstance(energies,YamboElectronsDB) and not energies.EXPAND) or isinstance(energies,YamboQPDB)
        no_symmetries = (isinstance(energies,YamboElectronsDB) and energies.EXPAND)

        if use_symmetries:
            if isinstance(energies,YamboElectronsDB): eigs = energies.eigenvalues_ibz[0,:,self.start_band:self.mband]
            if isinstance(energies,YamboQPDB):        
                # Pick the same energy range used for the BSE in case the QP energy range is larger
                if   energies.nbands<self.mband-self.start_band:  raise ValueError("[ERROR] QP range less than BSE range!")
                elif energies.nbands==self.mband-self.start_band: eigs = energies.eigenvalues_qp # Assuming exact same range
                else:                                             eigs = energies.eigenvalues_qp[:,self.start_band:self.mband]

            kpoints = car_red(np.array([ k/lat.alat for k in lat.ibz_kpoints ]),lat.rlat) #IBZ reduced kpoints 

            # we need to select the weights in the IBZ
            ibz_weights = np.zeros([lat.ibz_nkpoints,self.mband-self.start_band]) 
            for ik_bz,ik_ibz in enumerate(lat.kpoints_indexes): ibz_weights[ik_ibz] = weights[ik_bz]        
            weights = ibz_weights

            # sym_red are the symmetries of the reciprocal lattice...
            # skw interp wants the symmetries of the direct lattice, so we use sym_rec_red
            # plus, we take the non t-revved ones
            if not lat.mag_syms:
                symrel = [sym for sym,trev in zip(lat.sym_rec_red,lat.time_rev_list) if trev==False ]
                trev_for_interp = lat.time_rev
            # Handle special case of mag_sys + trev (e.g. SOC + ferromagnet, etc)
            elif lat.time_rev:
                symrel = lat.sym_rec_red
                trev_for_interp=False

        elif no_symmetries:
            eigs = energies.eigenvalues[0,:,self.start_band:self.mband]
            kpoints = lat.red_kpoints
            symrel = [np.identity(3)]
            trev_for_interp = False
        else:
            raise ValueError("Energies argument must be an instance of YamboElectronsDB or YamboQPDB. Got %s"%(type(energies)))

        # Additional input parameters for the SKW interpolator
        na = np.newaxis
        cell = (lat.lat, lat.red_atomic_positions, lat.atomic_numbers)
        nelect = 0
        fermie = kwargs.pop('fermie',0)

        # Get dense kpoints along the path (controlled by path.intervals)
        kpoints_path =  path.get_klist()[:,:3]
        
        #interpolate energies
        skw = SkwInterpolator(lpratio,kpoints,eigs[na,:,:],fermie,nelect,cell,symrel,trev_for_interp,verbose=verbose)
        exc_energies = skw.interp_kpts(kpoints_path).eigens

        #interpolate weights
        skw = SkwInterpolator(lpratio,kpoints,weights[na,:,:],fermie,nelect,cell,symrel,trev_for_interp,verbose=verbose)
        exc_weights = skw.interp_kpts(kpoints_path).eigens

        # For the band plot (bandstructure object), we need to switch to cartesian coordinates
        path_car = get_path_car(red_car(path.kpoints,lat.rlat),path)
        kpoints_path = path_car.get_klist()[:,:3]

        #create band-structure object
        exc_bands = YambopyBandStructure(exc_energies[0],kpoints_path,kpath=path_car,weights=exc_weights[0],size=size,**kwargs)
        #shift top v_band to zero
        exc_bands.set_fermi(self.nvbands)

        return exc_bands

    def interpolate_transitions(self,energies,path,excitons,lpratio=5,f=None,size=1,verbose=True,**kwargs):
        """ Interpolate exciton bandstructure using SKW interpolation from Abipy
        """

        lattice = self.lattice
        cell = (lattice.lat, lattice.red_atomic_positions, lattice.atomic_numbers)
        nelect = 0
        # Here there is something strange...
        fermie = kwargs.pop('fermie',0)
        ##
        if not lattice.mag_syms:
            symrel = [sym for sym,trev in zip(lattice.sym_rec_red,lattice.time_rev_list) if trev==False ]
            trev_for_interp = lattice.time_rev
        # Handle special case of mag_sys + trev (e.g. SOC + ferromagnet, etc)
        elif lattice.time_rev:
            symrel = lattice.sym_rec_red
            trev_for_interp=False

        #vmin, vmax = self.unique_vbands[0], self.unique_vbands[1]
        #cmin, cmax = self.unique_cbands[0], self.unique_cbands[1]

        transitions = self.get_exciton_transitions(excitons)
        transitions = transitions[:,:,:]

        ibz_nkpoints = max(lattice.kpoints_indexes)+1
        kpoints = lattice.red_kpoints

        #map from bz -> ibz:
        ibz_transitions = np.zeros([ibz_nkpoints,self.nvbands,self.ncbands])
        ibz_kpoints = np.zeros([ibz_nkpoints,3])
        for idx_bz,idx_ibz in enumerate(lattice.kpoints_indexes):
            ibz_transitions[idx_ibz,:,:] = transitions[idx_bz,:,:] 
            ibz_kpoints[idx_ibz] = lattice.red_kpoints[idx_bz]

        #get eigenvalues along the path
        if isinstance(energies,YamboElectronsDB):
            ibz_energies = energies.eigenvalues[:,self.start_band:self.mband]
        elif isinstance(energies,YamboQPDB):
            ibz_energies = energies.eigenvalues_qp
        else:
            raise ValueError("Energies argument must be an instance of YamboElectronsDB or YamboQPDB. Got %s"%(type(energies)))

        #interpolate energies
        na = np.newaxis
        skw = SkwInterpolator(lpratio,ibz_kpoints,ibz_energies[na,:,:],fermie,nelect,cell,symrel,trev_for_interp,verbose=verbose)
        kpoints_path = path.get_klist()[:,:3]
        energies = skw.interp_kpts(kpoints_path).eigens
     
        #interpolate transitions
        na = np.newaxis
        skw = SkwInterpolator(lpratio,ibz_kpoints,ibz_transitions[na,:,:],fermie,nelect,cell,symrel,trev_for_interp,verbose=verbose)
        kpoints_path = path.get_klist()[:,:3]
        exc_transitions = skw.interp_kpts(kpoints_path).eigens

        #create band-structure object
        exc_bands = YambopyBandStructure(energies[0],kpoints_path,kpath=path,weights=exc_weights[0],size=size,**kwargs)
        exc_bands.set_fermi(self.nvbands)

        return exc_transitions

    def interpolate_spin(self,energies,spin_proj,path,excitons,lpratio=5,f=None,size=1,verbose=True,**kwargs):
        """ Interpolate exciton bandstructure using SKW interpolation from Abipy
        """

        if verbose:
            print("This interpolation is provided by the SKW interpolator implemented in Abipy")

        lattice = self.lattice
        cell = (lattice.lat, lattice.red_atomic_positions, lattice.atomic_numbers)
        nelect = 0
        # Here there is something strange...

        fermie = kwargs.pop('fermie',0)
        ##
        symrel = [sym for sym,trev in zip(lattice.sym_rec_red,lattice.time_rev_list) if trev==False ]
        time_rev = True
 
        weights = self.get_exciton_weights(excitons)
        weights = weights[:,self.start_band:self.mband]
        if f: weights = f(weights)
        size *= 1.0/np.max(weights)
        ibz_nkpoints = max(lattice.kpoints_indexes)+1
        kpoints = lattice.red_kpoints

        #map from bz -> ibz:
        print("ibz_nkpoints")
        print(ibz_nkpoints)
        print("weights.shape")
        print(weights.shape)
        print(self.unique_vbands)
        print(self.unique_cbands)
        v_1 = self.unique_vbands[ 0]
        v_2 = self.unique_cbands[-1] + 1
        #exit()
        ibz_weights = np.zeros([ibz_nkpoints,self.nbands])
        ibz_kpoints = np.zeros([ibz_nkpoints,3])
        ibz_spin    = np.zeros([ibz_nkpoints,self.nbands])
        for idx_bz,idx_ibz in enumerate(lattice.kpoints_indexes):
            ibz_weights[idx_ibz,:] = weights[idx_bz,:] 
            ibz_kpoints[idx_ibz]   = lattice.red_kpoints[idx_bz]
            ibz_spin[idx_ibz,:]    = spin_proj[idx_bz,v_1:v_2]
        #get eigenvalues along the path
        if isinstance(energies,YamboElectronsDB):
            ibz_energies = energies.eigenvalues[:,self.start_band:self.mband]
        elif isinstance(energies,YamboQPDB):
            ibz_energies = energies.eigenvalues_qp
        else:
            raise ValueError("Energies argument must be an instance of YamboElectronsDB or YamboQPDB. Got %s"%(type(energies)))

        #interpolate energies
        na = np.newaxis
        print("na")
        print(na)
        skw = SkwInterpolator(lpratio,ibz_kpoints,ibz_energies[na,:,:],fermie,nelect,cell,symrel,time_rev,verbose=verbose)
        kpoints_path = path.get_klist()[:,:3]
        energies = skw.interp_kpts(kpoints_path).eigens
     
        #interpolate weights
        na = np.newaxis
        skw = SkwInterpolator(lpratio,ibz_kpoints,ibz_weights[na,:,:],fermie,nelect,cell,symrel,time_rev,verbose=verbose)
        kpoints_path = path.get_klist()[:,:3]
        exc_weights = skw.interp_kpts(kpoints_path).eigens

        #interpolate spin projection
        na = np.newaxis
        print("na")
        print(na)
        skw = SkwInterpolator(lpratio,ibz_kpoints,ibz_spin[na,:,:],fermie,nelect,cell,symrel,time_rev,verbose=verbose)
        kpoints_path = path.get_klist()[:,:3]
        spin_inter   = skw.interp_kpts(kpoints_path).eigens
        print("spin_inter")
        print(spin_inter)

        #create band-structure object
        exc_bands = YambopyBandStructure(energies[0],kpoints_path,kpath=path,weights=exc_weights[0],spin_proj=spin_inter[0],size=size,**kwargs)
        exc_bands.set_fermi(self.nvbands)

        return exc_bands
 
    def get_amplitudes_phases(self,excitons=(0,),repx=list(range(1)),repy=list(range(1)),repz=list(range(1))):
        """ get the excitonic amplitudes and phases
        """
        if self.eigenvectors is None:
            raise ValueError('This database does not contain Excitonic states,'
                             'please re-run the yambo BSE calculation with the WRbsWF option in the input file.')
        if isinstance(excitons, int):
            excitons = (excitons,)
       
        car_kpoints = self.lattice.car_kpoints
        nkpoints = len(car_kpoints)
        print(nkpoints)
        amplitudes = np.zeros([nkpoints])
        phases     = np.zeros([nkpoints],dtype=np.complex64)
        for exciton in excitons:
            #the the eigenstate
            eivec = self.eigenvectors[exciton-1]
           
            total = 0
            for eh,kvc in enumerate(self.table):
                ikbz, v, c = kvc-1
                Acvk = eivec[eh]
                phases[ikbz]     += Acvk
                amplitudes[ikbz] += np.abs(Acvk)

        #replicate kmesh
        red_kmesh,kindx = replicate_red_kmesh(self.lattice.red_kpoints,repx=repx,repy=repy,repz=repz)
        car_kpoints = red_car(red_kmesh,self.lattice.rlat)

        return car_kpoints, amplitudes[kindx], np.angle(phases)[kindx]

    def get_chi(self,emin=0,emax=10,estep=0.01,broad=0.1,q0norm=1e-5, nexcitons='all',spin_degen=2,verbose=0,**kwargs):
        """
        Calculate the BSE dielectric response function
        """
        if nexcitons == 'all': nexcitons = self.nexcitons

        #energy range
        w = np.arange(emin,emax,estep,dtype=np.float32)
        nenergies = len(w)
        
        if verbose:
            print("energy range: %lf -> +%lf -> %lf eV"%(emin,estep,emax))
            print("energy steps: %lf"%nenergies)
            print("broadening: %lf eV"%broad)

        #initialize the susceptibility intensity
        chi = np.zeros_like(w,dtype=np.complex64)

        # Oscillator strengths (residuals)
        EL1 = self.l_residual
        EL2 = self.r_residual

        # Broadening
        if isinstance(broad, float): # fixed broadening
            broad = np.full(nexcitons, broad) 
        elif isinstance(broad, tuple):  # linearly varying broadening
            broad_slope = broad[1] - broad[0]
            min_exciton = np.min(self.eigenvalues.real)
            broad = broad[0] + (self.eigenvalues[:nexcitons].real - min_exciton) * broad_slope

        # Excitonic states
        es = self.eigenvalues[:nexcitons, None]  # (nexcitons, 1)
        broad = broad[:, None]  # (nexcitons, 1)

        # Resonant and antiresonant GF structure
        G1 = -1 / (w - es + broad * I)  # (nexcitons, nenergies)
        G2 = -1 / (-w - es - broad * I)  # (nexcitons, nenergies)

        # chi = sum_s [ |R_s|^2/(w-E+i*eta) + |R_s|^2/(-w-E-i*eta) ]
        chi = np.einsum('s,sn->n', EL1 * EL2, G1 + G2)
        
        #dimensional factors
        try:
            if not self.Qpt=='1': q0norm = 2*np.pi*np.linalg.norm(self.car_qpoint)
        except:
            print("[WARNING] 1/q^2 set to 1 in eps2")
            q0norm=1
        try:
            if self.q_cutoff is not None: q0norm = self.q_cutoff
        except:
            print("[WARNING] 1/q^2 set to 1 in eps2")
            q0norm=1

        d3k_factor = self.lattice.rlat_vol/self.lattice.nkpoints
        cofactor = ha2ev*spin_degen/(2*np.pi)**3 * d3k_factor * (4*np.pi)  / q0norm**2

        chi = 1. + chi*cofactor #We are actually computing the epsilon, not the chi.

        return w,chi
    
    def get_pl(self,dipoles=None,dir=0,emin=0,emax=10,estep=0.01,broad=0.1,q0norm=1e-5, nexcitons='all',spin_degen=2,verbose=0,Boltz_Temp=300,**kwargs):
        """
        Calculate PL_0  using excitonic states
        """
        SPEED_OF_LIGHT    =  137*0.529*27.21/(6.582119569e-16)# finestructureconst * bohr2ang*HartreetoeV/hbar in eVs
        #SPEED_OF_LIGHT = 2.99792458e8
        RL_vol = self.lattice.rlat_vol
        nqbz =1 # For Gamma only calculation this should be ok, not sure to be checked
        # All prefactors should be checked
        pl_prefactor = 32.0*np.pi**3*SPEED_OF_LIGHT*2.0/3.0*RL_vol/nqbz/(2.00*np.pi)**3
        if nexcitons == 'all': nexcitons = self.nexcitons

        #energy range
        w = np.arange(emin,emax,estep,dtype=np.float32)
        nenergies = len(w)
        
        if verbose:
            print("energy range: %lf -> +%lf -> %lf "%(emin,estep,emax))
            print("energy steps: %lf"%nenergies)

        #initialize the susceptibility intensity
        pl = np.zeros([len(w)],dtype=np.complex64)

        if dipoles is None:
            #get dipole
            EL1 = self.l_residual
            EL2 = self.r_residual
        else:
            #calculate exciton-light coupling
            if verbose: print("calculate exciton-light coupling")
            EL1,EL2 = self.project1(dipoles.dipoles[:,dir],nexcitons) 

        if isinstance(broad,float): broad = [broad]*nexcitons

        if isinstance(broad,tuple): 
            broad_slope = broad[1]-broad[0]
            min_exciton = np.min(self.eigenvalues.real)
            broad = [ broad[0]+(es-min_exciton)*broad_slope for es in self.eigenvalues[:nexcitons].real]

        if "gaussian" in broad or "lorentzian" in broad:
            i = broad.find(":")
            if i != -1:
                value, eunit = broad[i+1:].split()
                if eunit == "eV": sigma = float(value)
                else: raise ValueError('Unknown unit %s'%eunit)

            f = gaussian if "gaussian" in broad else lorentzian
            broad = np.zeros([nexcitons])
            for s in range(nexcitons):
                es = self.eigenvalues[s].real
                broad += f(self.eigenvalues.real,es,sigma)
            broad = 0.1*broad/nexcitons

        #iterate over the excitonic states
        for s in range(nexcitons):
            #get exciton energy
            es = self.eigenvalues[s]
            es_0 = self.eigenvalues[0] # first exciton level
            print(es_0)
            pl_0_weight = boltzman_f(es-es_0, Boltz_Temp)
            #calculate the green's functions
            G1 = -1/(   w - es + broad[s]*I)
            G2 = -1/( - w - es - broad[s]*I)

            r = EL1[s]*EL2[s]*pl_0_weight#*pl_prefactor
            pl += r*G1 + r*G2

        #dimensional factors
        if not self.Qpt=='1': q0norm = 2*np.pi*np.linalg.norm(self.car_qpoint)
        if self.q_cutoff is not None: q0norm = self.q_cutoff

        d3k_factor = self.lattice.rlat_vol/self.lattice.nkpoints
        cofactor = ha2ev*spin_degen/(2*np.pi)**3 * d3k_factor * (4*np.pi)  / q0norm**2
        
        pl = 1. + pl*cofactor #We are actually computing the epsilon, not the chi.

        return w,pl

    def plot_chi_ax(self,ax,reim='im',n_brightest=-1,is_2D=False,**kwargs):
        """Plot chi on a matplotlib axes"""
        w,chi = self.get_chi(**kwargs)
        ## WARNING: assuming 
        ## (i)  nonperiodic direction is z 
        ## (ii) atomic units for ylat.lat
        if is_2D: 
            abs_label = 'alpha'
            Lz = self.lattice.lat[2,2] # interlayer separation in bohr
            chi = (chi-1.)*Lz/(4.*np.pi)
        else:
            abs_label = 'epsilon'
        #cleanup kwargs variables
        cleanup_vars = ['dipoles','dir','emin','emax','estep','broad',
                        'q0norm','nexcitons','spin_degen','verbose']
        for var in cleanup_vars: kwargs.pop(var,None)
        if 're' in reim: 
            ax.plot(w,chi.real,**kwargs)
            ax.set_ylabel(r'$Re(\%s(\omega))$' % abs_label)
        if 'im' in reim:
            ax.plot(w,chi.imag,**kwargs)
            ax.set_ylabel(r'$Im(\%s(\omega))$' % abs_label)
        ax.set_xlabel('Energy (eV)')
        #plot vertical bar on the brightest excitons
        if n_brightest>-1:
            exc_e,exc_i = self.get_sorted()
            for i,idx in exc_i[:n_brightest]:
                exciton_energy,idx = exc_e[idx]
                ax.axvline(exciton_energy,c='k')
                ax.text(exciton_energy,0.1,idx,rotation=90)
        return w,chi

    @add_fig_kwargs
    def plot_chi(self,n_brightest=-1,**kwargs):
        """Produce a figure with chi"""
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(1,1,1)
        self.plot_chi_ax(ax,n_brightest=n_brightest,**kwargs)
        return fig

    def save_chi(self,filename,**kwargs):
        """Compute chi and dump it to file"""
        w,chi = self.get_chi(**kwargs)
        np.savetxt(filename,np.array([w,chi.imag,chi.real]).T)

    ##########################################
    #  SPIN DEPENDENT PART UNDER DEVELOPMENT #
    ##########################################

    def exciton_bs_spin_pol(self,energies,path,excitons=(0,),debug=False):
        """
        Calculate exciton band-structure
        This function should contains the case of non-polarized calculations.
        Now is a first version
            
            Arguments:
            energies -> can be an instance of YamboElectronsDB or YamboQBDB
            path     -> path in reduced coordinates in which to plot the band structure
            exciton  -> exciton index to plot
        """
        if self.eigenvectors is None:
            raise ValueError('This database does not contain Excitonic states,'
                              'please re-run the yambo BSE calculation with the WRbsWF option in the input file.')
        if isinstance(excitons, int):
            excitons = (excitons,)

        #THESE THREE LINES FROM THE NOPOL CASE
        car_kpoints = self.lattice.car_kpoints
        rlat        = self.lattice.rlat
        bands_kpoints, band_indexes, path_car = get_path(car_kpoints,rlat,None,path,debug=debug) # None means the kpoints are already expanded
        
        if debug:
            for i,k in zip(band_indexes,band_kpoints):
                x,y,z = k
                plt.text(x,y,i) 
            plt.scatter(kpoints_rep[:,0],kpoints_rep[:,1])
            plt.plot(path[:,0],path[:,1],c='r')
            plt.scatter(band_kpoints[:,0],band_kpoints[:,1])
            plt.show()
            exit()

        #get eigenvalues along the path
        if isinstance(energies,YamboElectronsDB):
            #expand eigenvalues to the full brillouin zone
            if not energies.EXPAND: energies.expandEigenvalues()
            energies_up = energies.eigenvalues[0,:,:] # SPIN-UP
#            energies_up = energies.eigenvalues[0] # SPIN-UP
            energies_dw = energies.eigenvalues[1] # SPIN-DN 
            
        elif isinstance(energies,YamboQPDB):
            #expand the quasiparticle energies to the bull brillouin zone
            # Check this is OK
            # To-do
            print('todo')
            print(energies.eigenvalues_qp[self.lattice.kpoints_indexes].shape)
            pad_energies_up = energies.eigenvalues_qp[self.lattice.kpoints_indexes,:,0]
            pad_energies_dw = energies.eigenvalues_qp[self.lattice.kpoints_indexes,:,1]
            print(energies.max_band)
            min_band = energies.min_band       # check this
            print(pad_energies_up.shape)
            print(pad_energies_dw.shape)
            nkpoints, nbands = pad_energies_up.shape
            #exit()
            #pad_energies = energies.eigenvalues_qp[self.lattice.kpoints_indexes]
            #min_band = energies.min_band       # check this
            #nkpoints, nbands = pad_energies.shape
            energies_up = np.zeros([nkpoints,energies.max_band])
            energies_dw = np.zeros([nkpoints,energies.max_band])
            energies_up[:,min_band-1:] = pad_energies_up 
            energies_dw[:,min_band-1:] = pad_energies_dw
        else:
            raise ValueError("Energies argument must be an instance of YamboElectronsDB or YamboQPDB. Got %s"%(type(energies)))

        energies_up, energies_dw = energies_up[band_indexes],energies_dw[band_indexes]

        weights_up, weights_dw = self.get_exciton_weights_spin_pol(excitons)      
        weights_up, weights_dw = weights_up[band_indexes], weights_dw[band_indexes]

        #make top valence band to be zero
        fermi_level = max([ max(energies_up[:,max(self.unique_vbands)]), max(energies_dw[:,max(self.unique_vbands)] ) ])
        energies_up -= fermi_level  
        energies_dw -= fermi_level  
        
        return np.array(bands_kpoints), energies_up, energies_dw, weights_up, weights_dw, path_car

    def get_exciton_bs_spin_pol(self,energies_db,path,excitons,size_up=1,size_dw=1,space='bands',f=None,debug=False):
        """
        Get a YambopyBandstructure object with the exciton band-structure SPIN POLARIZED
        
            Arguments:
            ax          -> axis extance of matplotlib to add the plot to
            lattice     -> Lattice database
            energies_db -> Energies database, can be either a SaveDB or QPDB
            path        -> Path in the brillouin zone
        """
        from qepy.lattice import Path
        if not isinstance(path,Path): 
            raise ValueError('Path argument must be a instance of Path. Got %s instead'%type(path))
        if space == 'bands':
            if self.spin_pol=='pol':
               bands_kpoints, energies_up, energies_dw, weights_up, weights_dw, path_car = self.exciton_bs_spin_pol(energies_db, path, excitons, debug)
               nkpoints = len(bands_kpoints)
               plot_energies_up = energies_up[:,self.start_band:self.mband]
               plot_energies_dw = energies_dw[:,self.start_band:self.mband]
               plot_weights_up  = weights_up[:,self.start_band:self.mband]
               plot_weights_dw  = weights_dw[:,self.start_band:self.mband]
        else:
            raise NotImplementedError('TODO')
            eh_size = len(self.unique_vbands)*len(self.unique_cbands)
            nkpoints = len(bands_kpoints)
            plot_energies = np.zeros([nkpoints,eh_size])
            plot_weights = np.zeros([nkpoints,eh_size])
            for eh,(v,c) in enumerate(product(self.unique_vbands,self.unique_cbands)):
                plot_energies[:,eh] = energies[:,c]-energies[:,v]
                plot_weights[:,eh] = weights[:,c] 

        
        if f: plot_weights_up, plot_weights_dw = f(plot_weights_up), f(plot_weights_dw)
        size_plot_up = 100.0 # 1.0/np.max(plot_weights_up)*100.0
        size_plot_dw = 100.0 # 1.0/np.max(plot_weights_dw)*100.0
        ybs_up = YambopyBandStructure(plot_energies_up, bands_kpoints, weights=plot_weights_up, kpath=path_car, size=size_plot_up)
        ybs_dw = YambopyBandStructure(plot_energies_dw, bands_kpoints, weights=plot_weights_dw, kpath=path_car, size=size_plot_dw)

        return ybs_up, ybs_dw

    def get_exciton_weights_spin_pol(self,excitons):
    
        """get weight of state in each band for spin-polarized case"""
        table_up, table_dw, table_updw = [] , [], []
        for t,kcv in enumerate(self.table):
            k,c,v,c_s,v_s = kcv-1   # We substract 1 to be consistent with python numbering of arrays
            if c_s == 0 and v_s == 0:
               table_up.append(np.array(kcv[0:3]))
            if c_s == 1 and v_s == 1:
               table_dw.append(np.array(kcv[0:3]))
            if c_s == 1 and v_s == 0:
               table_updw.append(np.array(kcv[0:3]))
        table_up=np.array(table_up)
        table_dw=np.array(table_dw)
        table_updw=np.array(table_updw)

        self.unique_vbands_up = np.unique(table_up[:,1]-1)
        self.unique_cbands_up = np.unique(table_up[:,2]-1)
        self.unique_vbands_dw = np.unique(table_dw[:,1]-1)
        self.unique_cbands_dw = np.unique(table_dw[:,2]-1)
        self.mband_up = max(self.unique_cbands_up) + 1
        self.mband_dw = max(self.unique_cbands_dw) + 1
        self.start_band_up = min(self.unique_vbands_up)
        self.start_band_dw = min(self.unique_vbands_dw)

        weights_up = np.zeros([self.nkpoints,self.mband_up])
        weights_dw = np.zeros([self.nkpoints,self.mband_dw])
        
        for exciton in excitons:
            #get the eigenstate
            eivec = self.eigenvectors[exciton-1]
            #add weights
            sum_weights = 0
            for t,kcv in enumerate(self.table):
                k,c,v,c_s,v_s = kcv-1   # We substract 1 to be consistent with python numbering of arrays
                this_weight = abs2(eivec[t])
                if c_s == 0 and v_s == 0:
                   weights_up[k,c] += this_weight
                   weights_up[k,v] += this_weight
                elif c_s == 1 and v_s == 1: 
                   weights_dw[k,c] += this_weight
                   weights_dw[k,v] += this_weight
                sum_weights += this_weight
            if abs(sum_weights - 1) > 1e-3: raise ValueError('Excitonic weights does not sum to 1 but to %lf.'%sum_weights)
 
        return weights_up, weights_dw

    def interpolate_spin_pol(self,energies,path,excitons,lpratio=5,f=None,size_up=1.0,size_dw=1.0,verbose=True,**kwargs):
        """ Interpolate exciton bandstructure using SKW interpolation from
        Abipy and SPIN-POLARIZED CALCULATIONS

        AEK 14.08.2025
        It plots the weight on spin pol bands. 
        """
        if verbose:
            print("This interpolation is provided by the SKW interpolator implemented in Abipy")

        lattice = self.lattice
        cell = (lattice.lat, lattice.red_atomic_positions, lattice.atomic_numbers)
        nelect = 0
        # Here there is something strange...

        fermie = kwargs.pop('fermie',0)
        ##
        symrel = [sym for sym,trev in zip(lattice.sym_rec_red,lattice.time_rev_list) if trev==False ]
        time_rev = True
 
        weights_up, weights_dw = self.get_exciton_weights_spin_pol(excitons)

        weights_up = weights_up[:,self.start_band_up:self.mband_up]
        weights_dw = weights_dw[:,self.start_band_dw:self.mband_dw]

        if f: weights_up, weights_dw = f(weights_up), f(weights_dw)

        size_up *= 1.0/np.max(weights_up)
        size_dw *= 1.0/np.max(weights_dw)

        ibz_nkpoints = max(lattice.kpoints_indexes)+1
        kpoints = lattice.red_kpoints

        #map from bz -> ibz:
        # bug here? it is self.mband, but why?
        ibz_weights_up = np.zeros([ibz_nkpoints,self.mband_up-self.start_band_up]) 
        ibz_weights_dw = np.zeros([ibz_nkpoints,self.mband_dw-self.start_band_dw]) 
        
        ibz_kpoints = np.zeros([ibz_nkpoints,3])
        for idx_bz,idx_ibz in enumerate(lattice.kpoints_indexes):
#            print(weights_up[idx_bz,:])
            ibz_weights_up[idx_ibz,:], ibz_weights_dw[idx_ibz,:]= weights_up[idx_bz,:], weights_dw[idx_bz,:] 
            ibz_kpoints[idx_ibz] = lattice.red_kpoints[idx_bz]

        #get eigenvalues along the path
        # DFT values from SAVE
        if isinstance(energies,YamboElectronsDB):
            ibz_energies_up = energies.eigenvalues_ibz[0,:,self.start_band:self.mband] # spin-up channel
            ibz_energies_dw = energies.eigenvalues_ibz[1,:,self.start_band:self.mband] # spin-dw channel
            ibz_kpoints_qp  = ibz_kpoints
        # GW values from ndb.QP
        elif isinstance(energies,YamboQPDB):
            ibz_nkpoints_gw=len(energies.kpoints_iku)
            if not ibz_nkpoints == ibz_nkpoints_gw :
               print('GW and BSE k-grid are differents!')
               kpoints_gw_iku = energies.kpoints_iku
               kpoints_gw_car = np.array([ k/lattice.alat for k in kpoints_gw_iku ])
               kpoints_gw_red = car_red( kpoints_gw_car,lattice.rlat)
               ibz_kpoints_qp = kpoints_gw_red
            else:
               ibz_kpoints_qp = ibz_kpoints
            pad_energies_up = energies.eigenvalues_qp[:,:,0]
            pad_energies_dw = energies.eigenvalues_qp[:,:,1]
            min_band = energies.min_band
            nkpoints, nbands = pad_energies_up.shape

            ibz_energies_up = pad_energies_up 
            ibz_energies_dw = pad_energies_dw
            #print('ibz',ibz_energies_up.shape)
        else:
            raise ValueError("Energies argument must be an instance of YamboElectronsDB or YamboQPDB. Got %s"%(type(energies)))

        #interpolate energies
        na = np.newaxis

        skw_up = SkwInterpolator(lpratio,ibz_kpoints_qp,ibz_energies_up[na,:,:],fermie,nelect,cell,symrel,time_rev,verbose=verbose)
        skw_dw = SkwInterpolator(lpratio,ibz_kpoints_qp,ibz_energies_dw[na,:,:],fermie,nelect,cell,symrel,time_rev,verbose=verbose)
        kpoints_path = path.get_klist()[:,:3]
        energies_up = skw_up.interp_kpts(kpoints_path).eigens
        energies_dw = skw_dw.interp_kpts(kpoints_path).eigens
     
        #interpolate weights
        na = np.newaxis
        skw_up = SkwInterpolator(lpratio,ibz_kpoints,ibz_weights_up[na,:,:],fermie,nelect,cell,symrel,time_rev,verbose=verbose)
        skw_dw = SkwInterpolator(lpratio,ibz_kpoints,ibz_weights_dw[na,:,:],fermie,nelect,cell,symrel,time_rev,verbose=verbose)
        kpoints_path = path.get_klist()[:,:3]
        exc_weights_up = skw_up.interp_kpts(kpoints_path).eigens
        exc_weights_dw = skw_dw.interp_kpts(kpoints_path).eigens

        # Find and set the up-dw Fermi energy to zero
        self.nvbands_up = len(self.unique_vbands_up)
        self.nvbands_dw = len(self.unique_vbands_dw)
        fermi_up_dw = max([max(energies_up[0][:,self.nvbands_up-1]), max(energies_dw[0][:,self.nvbands_dw-1])])

        #create band-structure object
        kwargs.pop('size', None)
        exc_bands_up = YambopyBandStructure(energies_up[0],kpoints_path,kpath=path,fermie=fermi_up_dw,weights=exc_weights_up[0],size=size_up,**kwargs)
        exc_bands_dw = YambopyBandStructure(energies_dw[0],kpoints_path,kpath=path,fermie=fermi_up_dw,weights=exc_weights_dw[0],size=size_dw,**kwargs)
        return exc_bands_up, exc_bands_dw

    def get_magnon_bs(self,energies_db,path,excitons,size=1,space='bands',f=None,debug=False):
        #UNDER DEVELOPMENT
        """
        Get a YambopyBandstructure object with the exciton band-structure

            Arguments:
            ax          -> axis extance of matplotlib to add the plot to
            lattice     -> Lattice database
            energies_db -> Energies database, can be either a SaveDB or QPDB
            path        -> Path in the brillouin zone

            FP: to be moved in a separate class

            TO BE IMPLEMENTED
        """
        from qepy.lattice import Path
        if not isinstance(path,Path):
            raise ValueError('Path argument must be a instance of Path. Got %s instead'%type(path))

        if space == 'bands':
            bands_kpoints, energies, weights = self.magnon_bs(energies_db, path.kpoints, excitons, debug)
            nkpoints = len(bands_kpoints)
            plot_energies = energies[:,self.start_band:self.mband]
            plot_weights  = weights[:,self.start_band:self.mband]
        else:
            raise NotImplementedError('TODO')
            eh_size = len(self.unique_vbands)*len(self.unique_cbands)
            nkpoints = len(bands_kpoints)
            plot_energies = np.zeros([nkpoints,eh_size])
            plot_weights = np.zeros([nkpoints,eh_size])
            for eh,(v,c) in enumerate(product(self.unique_vbands,self.unique_cbands)):
                plot_energies[:,eh] = energies[:,c]-energies[:,v]
                plot_weights[:,eh] = weights[:,c]

        if f: plot_weights = f(plot_weights)
        size *= 1.0/np.max(plot_weights)
        ybs = YambopyBandStructure(plot_energies, bands_kpoints, weights=plot_weights, kpath=path, size=size)
        return ybs

    ##############################################
    #  END SPIN DEPENDENT PART UNDER DEVELOPMENT #
    ##############################################

    def get_string(self,mark="="):
        lines = []; app = lines.append
        app( marquee(self.__class__.__name__,mark=mark) )
        app( "BSE solved at Q:            %s"%self.Qpt )
        app( "number of excitons:         %d"%self.nexcitons )
        if self.Lkind is not None:
            app("L kind:                     %s"%self.Lkind)
        if self.table is not None: 
            app( "number of transitions:      %d"%self.ntransitions )
            app( "number of kpoints:          %d"%self.nkpoints  )
            app( "number of valence bands:    %d"%self.nvbands )
            app( "number of conduction bands: %d"%self.ncbands )
        return '\n'.join(lines)
    
    def __str__(self):
        return self.get_string()



from qepy import Path,PwXML
from math import sqrt

# k-points map
npoints = 50
path_kpoints = Path([ [[0.0, 0.0, 0.0],'$\Gamma$'],
                      [[0.5, 0.0, 0.0],'M'],
                      [[1./3,1./3,0.0],'K'],
                      [[0.0, 0.0, 0.0],'$\Gamma$']], [int(npoints*2),int(npoints),int(sqrt(5)*npoints)])

atom_1 = [range(16)]
atom_2 = [range(16,32)]

# Class PwXML. QE database reading
xml = PwXML(prefix='bn',path='bands')

# Class PwXML. QE database reading
xml.plot_eigen(path_kpoints)

# Alternative option with more plot control
## Matplotlib options 
#fig = plt.figure()
#ax = fig.add_subplot(1,1,1)

## Class PwXML. QE database reading
#xml.plot_eigen_ax(ax,path_kpoints)
#plt.show()



"""
Tutorial for the calculation of exciton radiative lifetimes in an ANISOTROPIC 3D bulk.
The tutorial can be adapted easily to the computation of exciton radiative lifetimes for 3D isotropic materials, 2D, 1D and 0D systems.


Plotting the radiative lifetime of the lowest-energy exciton of an ANISOTROPIC 3D material.
Plotting the thermal average exciton radiative lifetime of an ANISOTROPIC 3D material.

"""
from yambopy import YamboLatticeDB,YamboExcitonDB,YamboDipolesDB
from yambopy.bse.excitonradiativelifetimes import *
import numpy as np
import matplotlib.pyplot as plt


# modify paths accordingly 
lattice_path='./'
bse_path ='./'
dipoles_path ='./'

if __name__ == "__main__":

    #                    #
    # Start Yambopy part #
    #                    #

    # Create "lattice" object by reading the ns.db1 database
    ylat = YamboLatticeDB.from_db_file(filename=lattice_path+'/ns.db1')
    # Read exciton data
    yexc = YamboExcitonDB.from_db_file(ylat,filename=bse_path+'/ndb.BS_diago_Q1')
    # Read dipoles
    ydip = YamboDipolesDB.from_db_file(ylat,filename=dipoles_path+'/ndb.dipoles',project=False)


    # Provide the exciton effective masses and dielectric constants.
    #
    # Consider a uniaxial (ANISOTROPIC) 3D bulk material, with in-plane (xy) and out-of-plane (y) anisotropy
    #
    # Exciton effective masses in units of the electron rest mass
    M_xy = 2.5796  # m_e units
    M_z = 2.8778   # m_e units

    # Dielectric constants
    eps_xy = 6.852
    eps_z = 7.005

    # Define the temperature range

    Trange = np.arange(10,273,1)


    # Get the radiative lifetime of exciton state=0 in seconds
    state = 0
    tau_0 = get_radiative_lifetime_3D_aniso(Trange,state,ylat,ydip,yexc,[M_xy,M_z],[eps_xy,eps_z]) # seconds


    fig,ax = plt.subplots()
    ax.set_yscale('log')
    ax.set_xlabel('T (K)')
    ax.set_ylabel(r'$\tau_{S=0}$ (s)')
    ax.plot(Trange,tau_0)
    ax.set_title('Radiative lifetime of exciton state $S=0$')

    fig.savefig('exciton_0_radiative_lifetime.png',dpi=200,bbox_inches='tight')
    plt.show()


    # Get the thermal average radiative lifetime in seconds
    #
    # Define the exciton states included in the average
    states = range(10)
    tau_avg = average_lifetime(Trange,states,ylat,ydip,yexc,'3D_aniso',[M_xy,M_z],[eps_xy,eps_z])  # seconds

    fig,ax = plt.subplots()
    ax.set_yscale('log')
    ax.set_xlabel('T (K)')
    ax.set_ylabel(r'$\langle\tau\rangle$ (s)')
    ax.plot(Trange,tau_avg)
    ax.set_title('Average exciton radiative lifetime')

    fig.savefig('exciton_avg_radiative_lifetime.png',dpi=200,bbox_inches='tight')
    plt.show()



    #                  #
    # End Yambopy part #
    #                  #



#
# Authors: MN
#
import numpy as np

def find_degeneracy_evs(eigenvalues, atol=1e-3, rtol=1e-3):
    """
    Identify sets of degenerate eigenstates based on eigenvalues.

    Parameters
    ----------
    eigenvalues : array-like
        A list or array of eigenvalues
    atol : float, optional
        Absolute tolerance for degeneracy comparison. Default is 1e-3.
    rtol : float, optional
        Relative tolerance for degeneracy comparison. Default is 1e-3.

    Returns
    -------
    list of lists
        A list where each sublist contains the indices of degenerate eigenstates.

    Raises
    ------
    ValueError
        If `eigenvalues` is empty or not a valid array.
    """
    eigenvalues = np.asarray(eigenvalues)
    
    if eigenvalues.size == 0:
        raise ValueError("Input eigenvalues must not be empty.")
    if atol < 0 or rtol < 0:
        raise ValueError("Tolerances `atol` and `rtol` must be non-negative.")
    
    # Sort eigenvalues and get sorted indices
    idx_sorted = np.argsort(eigenvalues)
    eigenvalues_sorted = eigenvalues[idx_sorted]
    
    # Compute differences between consecutive eigenvalues
    diffs = np.diff(eigenvalues_sorted)
    tolerance = atol + rtol * np.abs(eigenvalues_sorted[:-1])
    
    # Identify where the differences exceed the tolerance
    split_indices = np.where(diffs > tolerance)[0]
    
    # Group indices of degenerate states (in sorted order)
    degen_sets_sorted = np.split(idx_sorted, split_indices + 1)

    return degen_sets_sorted


if __name__ == '__main__':
    eigenvalues = [3.0001, 2.99976, 1.99999, 1.0, 1.001, 1.002, 2.0, 2.001, 3.0]
    degen_sets = find_degeneracy_evs(eigenvalues, atol=1e-3, rtol=1e-3)
    print(degen_sets)



#
# Author: Henrique Pereira Coutada Miranda
# Run a BSE calculation using yambo
#
from __future__ import print_function
from builtins import range
import sys
from yambopy import *
from qepy import *
from schedulerpy import *
import argparse
import matplotlib.pyplot as plt

prefix = 'bn'
folder = 'bse'
yambo  = 'yambo'
p2y    = 'p2y'
ypp    = 'ypp'
layer_separation = 12
bash = Scheduler.factory

def create_save():
    #check if the nscf cycle is present
    if os.path.isdir('nscf/%s.save'%prefix):
        print('nscf calculation found!')
    else:
        print('nscf calculation not found!')
        exit()

    #check if the SAVE folder is present
    if not os.path.isdir('database/SAVE'):
        print('preparing yambo database')
        shell = bash()
        shell.add_command('mkdir -p database')
        shell.add_command('cd nscf/%s.save; %s; %s'%(prefix, p2y, yambo))
        shell.add_command('mv SAVE ../../database/')
        shell.run()
        shell.clean()

    if not os.path.islink('bse/SAVE'):
        s = bash()
        s.add_command('mkdir -p bse')
        s.add_command('cd bse; ln -s ../database/SAVE')
        if not dg: s.add_command('cd .. ; rm -f database/SAVE/ndb.Double_Grid')
        s.run()

    if dg and not os.path.isfile('database/SAVE/ndb.Double_Grid'):

        #check if the double grid nscf cycle is present
        if os.path.isdir('nscf_double/%s.save'%prefix):
            print('nscf_double calculation found!')
        else:
            print('nscf_double calculation not found!')
            exit()

        if not os.path.isdir('database_double/SAVE'):
            print('preparing yambo double database')
            shell = bash()
            shell.add_command('cd nscf_double/%s.save; %s; %s'%(prefix,p2y,yambo))
            shell.add_command('cd ../../')
            shell.add_command('mkdir -p database_double')
            shell.add_command('mv nscf_double/%s.save/SAVE database_double'%prefix)
            shell.run()

        #initialize the double grid
        print("creating double grid")

        yppin = YamboIn.from_runlevel('-m',filename='ypp.in',executable=ypp,folder='database')

        yppin['DbGd_DB1_paths'] = ["../database_double"]
        yppin.arguments.append('SkipCheck')

        yppin.write('database/ypp.in')

        shell = bash()
        shell.add_command('cd database; %s'%ypp)
        shell.add_command('cd ../%s ; rm -rf yambo o-*'%folder)
        #print(shell)
        shell.run()

def run(nthreads=1,cut=False):
    #create the folder to run the calculation
    if not os.path.isdir('bse'):
        shell = bash()
        shell.add_command('mkdir -p bse')
        shell.add_command('cp -r database/SAVE bse/')
        shell.run()
        shell.clean()

    #create the yambo input file
    y = YamboIn.from_runlevel('-r -d s -o b -k sex -y d -V all',executable=yambo,folder='bse')

    if cut:
        y['CUTGeo'] = 'box z'
        y['CUTBox'] = [0,0,layer_separation-1]

    y['FFTGvecs'] = [30,'Ry']
    y['NGsBlkXs'] = [1,'Ry']
    y['BndsRnXs'] = [1,70]
    y['BSEBands'] = [3,6]
    y['BEnSteps'] = 500
    y['BEnRange'] = [[0.0,10.0],'eV']
    """ 
    if os.path.isfile('gw/yambo/ndb.QP'):
        y['KfnQPdb'] = 'E < ../gw/yambo/ndb.QP' #Include previously computed quasiparticle energies
    else:
    """ 
    y['KfnQP_E']  = [2.91355133,1.0,1.0] #some scissor shift
    
    y.arguments.append('WRbsWF')

    if nthreads > 1:
        y['X_all_q_ROLEs'] = "q"
        y['X_all_q_CPUs'] = "%d"%nthreads

    y.write('bse/yambo_run.in')

    print('running yambo')
    shell = bash()
    if nthreads <= 1:
        shell.add_command('cd bse; %s -F yambo_run.in -J yambo'%yambo)
    else:
        shell.add_command('cd bse; mpirun -np %d %s -F yambo_run.in -J yambo'%(nthreads,yambo))
    if dg: shell.add_command('cd ..; rm database/SAVE/ndb.Double_Grid')
    shell.run()

if __name__ == "__main__":

    #parse options
    parser = argparse.ArgumentParser(description='Run BSE calculations on BN.')
    parser.add_argument('-dg','--doublegrid', action="store_true", help='Use double grid')
    parser.add_argument('-r', '--run',        action="store_true", help='Run BSE calculation')
    parser.add_argument('-c', '--cut',        action="store_true", help='Use coulomb truncation')
    parser.add_argument('-t' ,'--nthreads',                        help='Number of threads', default=1)
    args = parser.parse_args()
    nthreads = int(args.nthreads)

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    cut = args.cut
    dg = args.doublegrid

    create_save()
    if args.run:     run(nthreads,cut)



# Copyright (c) 2023-2026, Claudio Attaccalite,
#                          Myrta Gruening
# All rights reserved.
#
# This file is part of the yambopy project
# Calculate linear response from real-time calculations (yambo_nl)
# Modified to allow generalisation
#
import numpy as np
from yambopy.units import ha2ev,fs2aut, Junit, EFunit,SVCMm12VMm1,AU2VMm1
from yambopy.nl.external_efield import Divide_by_the_Field
from tqdm import tqdm
from scipy.optimize import least_squares
import sys
import os
from abc import ABC,abstractmethod
#
#
# Template class
#
class Xn_from_signal(ABC):
    def __init__(self,nldb,X_order=4,T_range=[-1, -1],l_out_current=False,nsamp=-1,samp_mod='',solver='',tol=1e-10,debug_mode=False): 
        self.time = nldb.IO_TIME_points # Time series
        self.T_step =self.time[1] - self.time[0]     # Time step of the simulation
        self.T_deph =12.0/nldb.NL_damping 
        self.efield = nldb.Efield[0] # External field of the first run
        self.pumps = nldb.Efield_general[1:] 
        self.efields = nldb.Efield
        self.n_runs = len(nldb.Polarization)     # Number of external laser frequencies
        self.polarization = nldb.Polarization     # Array of polarizations for each laser frequency
        self.current     =nldb.Current # Array of currents for each laser frequency
        self.l_eval_current=nldb.l_eval_CURRENT
        self.l_out_current = l_out_current and self.l_eval_current
        self.X_order = X_order
        SOLVE_MODES = ['', 'full', 'lstsq', 'lstsq_opt', 'svd']
        if solver not in SOLVE_MODES: 
            raise ValueError("Invalid solver mode. Expected one of: %s" % SOLVE_MODES)
        self.solver = solver
        SAMP_MODES = {'','linear', 'log', 'random'}    
        if samp_mod not in SAMP_MODES:
            raise ValueError(f"Invalid sampling mode. Expected one of: {SAMP_MODES}")
        self.samp_mod = samp_mod
        self.nsamp = nsamp
        self.T_urange = T_range 
        self.freqs = np.array([efield["freq_range"][0] for efield in nldb.Efield], dtype=np.double)     # Harmonic frequencies
        self.prefix = f'-{nldb.calc}' if nldb.calc != 'SAVE' else ''
        self.out_dim = 0
        self.tol = tol
        self.debug_mode = debug_mode
        super().__init__()
        
    def __str__(self): 
        """
        Print info of the class
        """
        s="\n * * *  Xn from signal class  * * * \n\n"
        s+="Max time: "+str(self.time[-1])+"\n"
        s+="Time step : "+str(self.T_step)+"\n"
        s+="Type Efield    : "+str(self.efield["name"])+"\n"
        s+="Number of runs   : "+str(self.n_runs)+"\n"
        s+="Current evaluated    : "+str(self.l_eval_current)+"\n"
        s+="Max harmonic order   : "+str(self.X_order)+"\n"
        if self.samp_mod != '':
            s+="Sampling mode    : "+str(self.samp_mode)+"\n"
        if self.solver != '':
            s+="Solver           : "+str(self.solver)+"\n"
        if self.nsamp > 0:
            s+="Sampling points  : "+str(self.nsamp)+"\n"
        if self.T_urange!=[-1, -1]:
            s+="User time range      : "+str(self.T_urange)+" [au] \n"
        s+="Frequency range: ["+str(self.freqs[0])+","+str(self.freqs[-1])+"] [au] \n"
        if  self.l_out_current:
            s+="Output is set to current."
        else:
            s+="Output is set to polarization."
        return s
    
    @abstractmethod
    def set_defaults(self):
        pass

    @abstractmethod
    def set_sampling(self,ifrq): 
       pass
   
    @abstractmethod
    def define_matrix(self,samp,ifrq):
        pass

    @abstractmethod
    def update_time_range(self):
        pass

    @abstractmethod
    def get_Unit_of_Measure(self,i_order):
        pass

    def get_sampling(self,T_range,idir,ifrq): 
        i_t_start = int(np.round( T_range[0] / self.T_step)) 
        i_deltaT  = int(np.round((T_range[1]-T_range[0])/self.T_step)/self.nsamp)
        if self.samp_mod=='linear':
            i_t = i_t_start + i_deltaT * np.arange(self.nsamp)
        elif self.samp_mod=='log':
            T_i = np.geomspace(i_t_start * self.T_step, T_range[1], self.nsamp, endpoint=False)
            i_t = np.array(np.round(T_i/self.T_step),dtype=int)
        elif self.samp_mod=='random':
            i_t = np.random.uniform(i_t_start,int(np.round(t / self.T_step)), self.nsamp)
        T_i = i_t*self.T_step - self.efield["initial_time"] 
        if self.l_out_current:
            S_i = np.array([self.current[ifrq][idir,i] for i in i_t])
        else:
            S_i = np.array([self.polarization[ifrq][idir,i] for i in i_t]) 
        return T_i,S_i
    
    def solve_lin_system(self,mat,samp,init=None):
        mat_dim = int(mat.shape[1])
        out=np.zeros(mat_dim,dtype=np.cdouble)
        if self.solver=="full" and ((not self.IsSquare(mat)) or (not self.IsWellConditioned(mat))):
            print(f'WARNING: solver changed to least square since square:{self.IsSquare(mat)} and well-conditioned:{self.IsWellConditioned(mat)}')
            self.solver = "lstsq"
        if self.solver=="full":
            out = np.linalg.solve(mat,samp)
        if self.solver=="lstsq":
            out = np.linalg.lstsq(mat,samp,rcond=self.tol)[0]
        if self.solver=="svd":
            inv = np.linalg.pinv(mat,rcond=self.tol)
            for i_n in range(mat_dim):
                out[i_n]=out[i_n]+np.sum(inv[i_n,:]*samp[:])
        if self.solver=="lstsq_opt":
            if(init is None):
                x0_cmplx = np.linalg.lstsq(mat, samp, rcond=tol)[0]
            else:
                x0_cmplx = init
            x0 = np.concatenate((x0_cmplx.real, x0_cmplx.imag))
            res = least_squares(residuals_func, x0, ftol=1e-11,gtol=1e-11,xtol=1e-11,verbose=0,x_scale='jac',args=(mat,samp))
            out = res.x[0:int(res.x.size/2)] + 1j * res.x[int(res.x.size/2):res.x.size]
        return out
    
    def perform_analysis(self):
        _ = self.set_defaults()
        out = np.zeros((self.out_dim, self.n_runs, 3), dtype=np.cdouble) 
        for i_f in tqdm(range(self.n_runs)):
            T_r = self.set_sampling(i_f)
            for i_d in range(3):
                samp_time, samp_sig= self.get_sampling(T_r,i_d,i_f)
                matrix = self.define_matrix(samp_time,i_f)
                raw = self.solve_lin_system(matrix,samp_sig)
                out[:, i_f, i_d] = raw[:self.out_dim]
                if self.debug_mode:
                    print(f"Freq #{i_f}, direction: {i_d}:")
                    print("***Sampling:")
                    print(samp_time, samp_sig)
                    print("***Matrix:")
                    print(matrix)
                    print("***Solution:")
                    print(raw)
        return out

    def get_Unit_of_Measure(self,i_order): # not sure if this is a general or specific method - let it here for the moment
        linear = 1.0
        ratio = SVCMm12VMm1 / AU2VMm1 # From AU to statVolt/cm ...is there a better way than this?
        if self.l_out_current:
            linear = Junit/EFunit
            ratio = 1.0/EFunit
        if i_order == 0:
            return np.power(ratio, 1, dtype=np.float64)*linear
        return np.power(ratio, i_order - 1, dtype=np.float64)*linear

    @abstractmethod
    def output_analysis(self,out,to_file):
        pass

    @abstractmethod
    def reconstruct_signal(self,out,to_file):
        pass

    def append_runinfo(self,T):
        s = "# Field details:"
        s+= "# Type field            "+str(self.efield["name"])+"\n"
        s+= "# Field intensity       "+str(self.efield["intensity"])+"\n"
        s+= "# Field versor          "+str(self.efield["versor"])+"\n"
        if self.efield["name"]=='QSSIN':
            s+= "Centre Gaussian     "+str(self.T0/fs2aut)+"[fs] \n"
            s+= "Gaussian sigma      "+str(self.sigma/fs2aut)+"[fs] \n"
        s+= "# Analysis details:"
        s+="# Max harmonic order   : "+str(self.X_order)+"\n"
        s+="# Sampling mode        : "+str(self.samp_mod) +"\n"
        s+="# Solver               : "+str(self.solver)+"\n"
        s+="# Sampling points      : "+str(self.nsamp)+"\n"
        s+="# Start sampling time  : "+str(T/fs2aut)+" [fs] \n"
        return s
        

### some maths auxiliary functions
    def IsSquare(self,m):
        return m.shape[0] == m.shape[1]

    def IsWellConditioned(self,m): # with this I am trying to avoid inverting a matrix
        return np.linalg.cond(m) < 1/sys.float_info.epsilon

    def residuals_func(x,M,S_i):
        x_cmplx=x[0:int(x.size/2)] + 1j * x[int(x.size/2):x.size]
        return np.linalg.norm(np.dot(M, x_cmplx) - S_i)



from netCDF4 import Dataset
from yambopy import LetzElphElectronPhononDB,ConvertElectronPhononDB
import os
import shutil
import subprocess
"""
Calculate gauge-invariant electron-phonon matrix elements with LetzElPhC and convert them into Yambo format

- Usage:

>> yambopy l2y -ph phinp -b b1 b2 -par nq nk [--kernel kernel] [--lelphc lelphc] [--no_lelphc_dbs] [--no_gkkp]

- Input parameters:
	-ph                   : path to ph.x input file, e.g. dvscf/ph.in
	-b                    : initial and final band indices (counting from 1, both values included)
	-par [OPT]            : MPI pools for q and k (needs mpirun)
	--kernel [OPT]        : e-ph kernel type, default 'dfpt'
	--lelphc [OPT]        : path to lelphc executable, default 'lelphc', code will prompt
	--no_lelphc_dbs [OPT] : will remove LetzElPhC input and outputs
    --no_gkkp [OPT]       : do not generate the gkkp databases for Yambo/Lumen
 
- Prerequisites:

* ph.x phonon calculation must be complete, e.g. the phinp folder should contain:
    - ph.x input file
    - pw.x (scf) save directory
    - [prefix].dyn files
    - _ph* directories
* Yambo SAVE directory must be present. We run in the directory where the SAVE is.
* LetzElPhC must be installed
* mpirun must be linked for parallel runs
"""

def get_input(bands,kernel,pools,ph_path):
	"""
	Get LetzElPhC input
	"""
	input_file = []; app = input_file.append
	app("# LetzElPhC input for yambo generated by yambopy")
	app(f"nqpool      = {pools[0]}") # first pool is qpoints
	app(f"nkpool      = {pools[1]}") # second pool is kpoints
	app(f"start_bnd   = {bands[0]}")
	app(f"end_bnd     = {bands[1]}")
	app(f"save_dir    = ./SAVE")
	app(f"kernel      = {kernel}")
	app(f"ph_save_dir = {ph_path}/ph_save")
	app(f"convention = yambo")
	return "\n".join(input_file)

def checks(phinp,lelphc,bands,kernel,pools):

	## ph.x input file and directory
	if '/' in phinp:
		path_ph = phinp.rsplit('/',1)[0]
		inp_ph  = phinp.rsplit('/',1)[1]
	else:
		path_ph='./'
		inp_ph = phinp

	## check for SAVE
	if not os.path.isdir("SAVE"):
		raise FileNotFoundError("SAVE directory not found. Make sure to run in the directory where the SAVE is.")

	## check for lelphc executable
	is_lelphc = shutil.which(lelphc)
	if is_lelphc is None: lelphc=input("lelphc executable not found. Please provide absolute path here: \n")

	## check for mpirun
	if int(pools[0])*int(pools[1])>1:
		is_mpirun = shutil.which('mpirun')
		if is_mpirun is None: 
			print("[WARNING] mpirun not found, running in serial") 
			pools = [1,1]

	## check band indices
	try: assert(int(bands[0])<int(bands[1]))
	except: raise ValueError("[ERROR] band indices must be integers with b1<b2")

	## check kernel type
	if kernel not in ['dfpt','bare','dfpt_local','bare_local']:
		raise ValueError(f"[ERROR] Unrecognized kernel type {kernel}")

	## lelphc input file
	inp_lelphc = get_input(bands,kernel,pools,path_ph)
	inp_name = 'lelphc.in'

	return lelphc,path_ph,inp_ph,inp_lelphc,inp_name

def run_preprocessing(lelphc,path_ph,inp_ph):

	print(":: LetzElPhC pre-processing ::")
	pp_run_str = f"{lelphc} -pp --code=qe -F {inp_ph}"
	try:
		pp=subprocess.run(f'cd {path_ph} ; {pp_run_str} ; cd -',shell=True,check=True,capture_output=True,text=True)
	except subprocess.CalledProcessError as e:
		print("PP Error:", e)
		print("Description", e.stderr)

	if not os.path.isdir(f"{path_ph}/ph_save"):
		print("[ERROR] Something wrong with preprocessing step:")
		print(pp.stderr)
		exit()

def run_elph(lelphc,inp_lelphc,inp_name):

	def get_ntasks(inp):
		ntasks=1
		for line in inp.split('\n'):
			if 'pool' in line: ntasks *= int(line.split()[-1])
		return ntasks			

	print(":: LetzElPhC el-ph calculation ($> tail -f lelphc.log) ::")

	# write input file
	f = open(inp_name, "w")
	f.write(inp_lelphc)
	f.close()

	# manage parallel run
	ntasks = get_ntasks(inp_lelphc)
	if ntasks > 1 : elph_run_str = f"mpirun -np {ntasks} {lelphc} -F {inp_name}"
	elif ntasks==1: elph_run_str = f"{lelphc} -F {inp_name}"
	else: raise ValueError(f'Wrong MPI task selection: {ntasks} tasks?')

	# run lelphc
	try:
		with open('lelphc.log','w') as log_file:
			elph_run=subprocess.run(elph_run_str,shell=True,check=True,capture_output=False,text=True,stdout=log_file)
	except subprocess.CalledProcessError as e:
		print("Error:", e)
		print("Description:", e.stderr)

def letzelph_to_yambo():

	print(":: Load el-ph database ::")
	lelph_obj = LetzElphElectronPhononDB('ndb.elph',div_by_energies=False)

	print(":: Convert el-ph data to yambo format ::")
	l2y       = ConvertElectronPhononDB(lelph_obj,'lelphc','SAVE','SAVE')

def clean_lelphc(no_lelphc_dbs,inp_name,ph_path):
	if not no_lelphc_dbs:
		print(":: LetzElPhC databases ::")
		print("     - lelphc.in input written")
		print("     - lelphc.log output written")
		print("     - ph_save | ndb.elph | ndb.Dmats written")
	else:
		[os.remove(C) for C in [inp_name,'ndb.elph','ndb.Dmats', 'lelphc.log'] if os.path.isfile(C)]
		if os.path.isdir(f"{ph_path}/ph_save"): shutil.rmtree(f"{ph_path}/ph_save")

if __name__=="__main__":

	parser = argparse.ArgumentParser(description='Generate electron-phonon coupling databases via LetzElPhC')
	parser.add_argument('-ph','--ph_inp_path', type=str, help='<Required> Path to ph.x (dvscf) input file',required=True)
	parser.add_argument('-b','--bands',nargs='2',type=str,help="<Required> First and last band (counting from 1), e.g. 'b_i b_f'",required=True)
	parser.add_argument('-k','--kernel', type=str, default='dfpt',help="<Optional> Electron-phonon kernel type, e.g. 'dfpt', 'bare', ... (default 'dfpt')")
	parser.add_argument('-par','--pools',nargs='2',type=str,default=[1,1],help="<Optional> MPI tasks as 'nqpools nkpools' (default serial)")
	parser.add_argument('-lelphc','--lelphc',type=str,default='lelphc',help="<Optional> Path to lelphc executable (default assumed in Path, otherwise prompted)")
	parser.add_argument('-nl','--no_lelphc_dbs', action="store_true", help="Remove lelphc databases (False if not given)")
	parser.add_argument('-ng','--no_gkkp', action="store_true", help="Do not generate gkkp dbs (False if not given)")
	args = parser.parse_args()

	phinp     = args.ph_inp_path
	bands     = args.bands
	kernel    = args.kernel
	pools     = args.pools
	lelphc    = args.lelphc
	no_lelphc = args.no_lelphc_dbs
	no_gkkp   = args.no_gkkp

	# Check inputs
	lelphc,ph_path,inp_ph,inp_lelphc,inp_name = checks(phinp,lelphc,bands,kernel,pools)

	# run preprocessing
	run_preprocessing(lelphc,ph_path,inp_ph)

	# run el-ph calculation and rotation
	run_elph(lelphc,inp_lelphc,inp_name,pools)

	# load database and convert to yambo format
	if not no_gkkp: letzelph_to_yambo()

	# clean
	clean_lelphc(no_lelphc_dbs,inp_name,ph_path)	



#
# Authors: MN
"""
Script to compute total crystal angular momentum
This script loads Yambo databases, calculates the angular momentum for a specific exciton, 
and generates a real-space visualization (.cube file).
"""

import numpy as np
from yambopy.dbs.excitondb import YamboExcitonDB
from yambopy.dbs import excitondb
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.dbs.wfdb import YamboWFDB
import os

## inputs 
iqpt = 1
# This is the index of the q-point (momentum transfer). 

path = '.'
# This defines the working directory where your 'GW_BSE' and 'SAVE' folders are located.
# '.' means the current directory.

gw_bse_dir = 'GW_BSE'
# BSE Job name. This is where your bse file are store after your yambo bse run.

iexe = 0
# This is the index of the specific exciton state you want to analyze (python indexing).
# You likely found this index by looking at the absorption spectrum or energy table previously.

degen_tol = 1e-2
# This is the energy tolerance in eV.
# States with energy differences smaller than this value are considered "degenerate" (the same energy)
# and will be treated together during the angular momentum mixing.

# ======= Load data ========

## First load the lattice db
lattice = YamboLatticeDB.from_db_file(os.path.join(path, 'SAVE', 'ns.db1'))
# We load the Lattice Database (ns.db1) from the SAVE folder.
# This file contains information about the crystal structure, unit cell geometry, and k-points.

# load exciton db
filename = 'ndb.BS_diago_Q%d' % (iqpt)
excdb = YamboExcitonDB.from_db_file(lattice, filename=filename,
                                    folder=os.path.join(path, gw_bse_dir), neigs = iexe + 101)
# We load the Exciton Database.
# Critical Note: 'neigs' (number of eigenvalues) is set to 'iexe + 100'. 
# We load more states than the target 'iexe' to ensure we capture all states that might be degenerate 
# with our target. If you don't load the degenerate partners, the physics will be wrong.

# Load the wavefunction database
wfdb = YamboWFDB(path=path, latdb=lattice, bands_range=[np.min(excdb.table[:, 1]) - 1, np.max(excdb.table[:, 2])])

symm_mat_cart = lattice.sym_car[2] 
# The rotation symmetry for which we want to find total_crys_angular_momentum values and simentineous eigenbasis
# For example, here I am selecting 3th symmetry matrix which corresponds to 120 degree rotation along z (I am doing this hBN).
# you can check this in qe scf out file when verbosity = 'high'
# Neverthesless you can also give your custom 3x3 matrix which must satisfy Rq = q and is a symmetry of the crystal.

frac_trans_cart = np.zeros(3)
# Fractional translation vector in cart units for the corre symmetry.
# i.e x -> Rx + t where R is symm_mat_cart and t is frac_trans_cart 

sbasis = excdb.total_crys_angular_momentum(wfdb, iexe, symm_mat_cart, frac_trans_cart, degen_tol=degen_tol)
# This is the core calculation.
# It computes the "Total Crystal Angular Momentum" for the target exciton ('iexe').
# It returns 'sbasis', which contains the angular momentum values and the new basis vectors.

jvals = sbasis[0]
# These are the total crystal angular momentum eigenvalues (the 'j' values) for the degenerate states.

Akcv_j = sbasis[1]
# These are the new right eigenvectors (Akcv) for the excitons i.e are simentineous eigenstates of Hbse and symmetry operator.
# i.e they represent how to mix the original states to create states with well-defined angular momentum.

print('Total crystal angular momentum : ', jvals)
# Prints the calculated j-values to the screen.

## Now we Plot read space wf
idegen = np.array(excdb.get_degenerate(iexe + 1, eps=degen_tol), dtype=int) - 1
# We find the indices of all states that are degenerate with our target exciton.
# 'iexe+1' is used because some internal functions use 1-based indexing, while python uses 0-based.
# The result 'idegen' is an array of Python indices (0-based) for these states.

## Update the existing wfs with the new rotated similentious eigenbasis
old_eig_states = excdb.Akcv[idegen].copy()
excdb.Akcv[idegen] = Akcv_j
# We overwrite the original exciton coefficients in the database object with our new, rotated coefficients.
# This ensures that when we plot the state later, we are plotting the angular momentum eigenstate, 
# not the arbitrary original state.

## we break degeneracies to plot each state rather than averaging. THis is to trick the real_wf_to_cube
## as it will always internally average the eigenstates 
degen_tol_add = excdb.eigenvalues.max()*4
# We prepare a large number to shift the energies.
# Plotting tools often average degenerate states. To plot a specific 'j' state individually,
# we need to trick the tool into thinking they have different energies.

old_eig_states_ene = excdb.eigenvalues[idegen].copy()
for i in idegen:
    degen_tol_add += 1 
    # Increment the shift value.
    excdb.eigenvalues[i] += degen_tol_add
    # We artificially shift the energy of this specific state.
    # Now this state is energetically isolated and won't be averaged with others during plotting.

# Now lets plot the real space wavefunction for each j value
print("="*80)
for i in range(len(jvals)):
    print('*********** Plotting real space wf for state %d with j = %.2f ***************'
          % (idegen[i] + 1, jvals[i]))
    print("="*80)
    excdb.real_wf_to_cube(iexe=idegen[i], wfdb=wfdb, fixed_position=[0.0, 0.0, 0.0],
                          supercell=[9,9,9], degen_tol=1e-14, wfcCutoffRy=-1,
                          fix_particle='e', phase=True)
    print("="*80)

## restore the db back
excdb.eigenvalues[idegen] = old_eig_states_ene
excdb.Akcv[idegen] = old_eig_states

## .cube will be dumped and use vesta to visualize it !
